from plone import api
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from zope.component import getMultiAdapter

import unittest


class TestSerializeUserToJsonAdapter(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        properties = {
            "email": "noam.chomsky@example.com",
            "username": "noamchomsky",
            "fullname": "Noam Avram Chomsky",
            "home_page": "web.mit.edu/chomsky",
            "description": "Professor of Linguistics",
            "location": "Cambridge, MA",
        }
        self.user = api.user.create(
            email="noam.chomsky@example.com", username="noam", properties=properties
        )

    def serialize(self, user):
        serializer = getMultiAdapter((user, self.request), ISerializeToJson)
        return serializer()

    def test_serialize_returns_id(self):
        user = self.serialize(self.user)
        self.assertTrue(user)
        self.assertEqual("noam", user.get("id"))
        self.assertEqual("noam.chomsky@example.com", user.get("email"))
        self.assertEqual("Noam Avram Chomsky", user.get("fullname"))
        self.assertEqual("web.mit.edu/chomsky", user.get("home_page"))  # noqa
        self.assertEqual("Professor of Linguistics", user.get("description"))  # noqa
        self.assertEqual("Cambridge, MA", user.get("location"))

    def test_serialize_roles(self):
        user = self.serialize(self.user)
        self.assertIn("roles", user)
        self.assertNotIn("Authenticated", user["roles"])
        self.assertNotIn("Anonymous", user["roles"])

    def test_serialize_custom_member_schema(self):
        from plone.app.users.browser.schemaeditor import applySchema

        member_schema = """
            <model xmlns="http://namespaces.plone.org/supermodel/schema"
                xmlns:form="http://namespaces.plone.org/supermodel/form"
                xmlns:users="http://namespaces.plone.org/supermodel/users"
                xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                i18n:domain="plone">
              <schema name="member-fields">
                <field name="twitter" type="zope.schema.TextLine"
                         users:forms="In User Profile">
                  <description i18n:translate="help_twitter">
                    Twitter account
                  </description>
                  <required>False</required>
                  <title i18n:translate="label_twitter">Twitter Account</title>
                </field>
              </schema>
            </model>
        """
        applySchema(member_schema)
        user = api.user.create(
            email="donald.duck@example.com",
            username="donald",
            properties={"twitter": "TheRealDuck"},
        )
        res = self.serialize(user)
        self.assertIn("twitter", res)
        self.assertEqual(res["twitter"], "TheRealDuck")



"""Utils to translate FTIs / zope.schema interfaces to JSON schemas.

The basic idea here is to instantiate a minimal z3c form, and then have
plone.autoform work its magic on it to process all the fields, and apply
any p.autoform directives (fieldsets, field modes, omitted fields, field
permissions, widgets).

Also schema interface inheritance (IRO) and additional schemata from behaviors
are factored into how the final resulting fieldsets are composed.

This approach should ensure that all these directives get respected and
processed the same way they would for a server-rendered form.
"""

from collections import OrderedDict
from copy import copy
from plone.autoform.form import AutoExtensibleForm
from plone.autoform.interfaces import IParameterizedWidget
from plone.autoform.interfaces import WIDGETS_KEY
from plone.behavior.interfaces import IBehavior
from plone.dexterity.interfaces import IDexterityContent
from plone.dexterity.interfaces import IDexterityFTI
from plone.dexterity.utils import getAdditionalSchemata
from plone.i18n.normalizer import idnormalizer
from plone.restapi.interfaces import IFieldDeserializer
from plone.restapi.serializer.converters import IJsonCompatible
from plone.restapi.types.interfaces import IJsonSchemaProvider
from plone.supermodel import serializeModel
from plone.supermodel.interfaces import FIELDSETS_KEY
from plone.supermodel.utils import mergedTaggedValueDict
from plone.supermodel.utils import syncSchema
from Products.CMFCore.utils import getToolByName
from z3c.form import form as z3c_form
from zExceptions import BadRequest
from zope.component import getMultiAdapter
from zope.component import queryMultiAdapter
from zope.component import queryUtility
from zope.component.hooks import getSite
from zope.globalrequest import getRequest
from zope.i18n import translate
from zope.interface import implementer
from zope.schema.interfaces import IVocabularyFactory
from plone.app.multilingual.dx.interfaces import MULTILINGUAL_KEY
from plone.supermodel.utils import mergedTaggedValueList

try:
    # Plone 5.1+
    from plone.dexterity.schema import splitSchemaName
except ImportError:
    # Plone 4.3
    from plone.dexterity.utils import splitSchemaName


_marker = []  # Create a new marker object.

FIELD_PROPERTIES_MAPPING = {
    "minLength": "min_length",
    "maxLength": "max_length",
    "minItems": "min_length",
    "maxItems": "max_length",
    "minimum": "min",
    "maximum": "max",
}


@implementer(IDexterityContent)
class FakeDXContext:
    """Fake DX content class, so we can re-use the DX field deserializers"""


def create_form(context, request, base_schema, additional_schemata=None):
    """Create a minimal, standalone z3c form and run the field processing
    logic of plone.autoform on it.
    """
    if additional_schemata is None:
        additional_schemata = ()

    class SchemaForm(AutoExtensibleForm, z3c_form.AddForm):
        schema = base_schema
        additionalSchemata = additional_schemata
        ignoreContext = True
        ignorePrefix = True

    form = SchemaForm(context, request)
    form.updateFieldsFromSchemata()
    return form


def iter_fields(fieldsets):
    """Iterate over a flat list of fields, given a list of fieldset dicts
    as returned by `get_fieldsets`.
    """
    for fieldset in fieldsets:
        yield from fieldset["fields"]


def get_form_fieldsets(form):
    """Get fieldsets from form"""
    fieldsets = []
    form_fields = getattr(form, "fields", {})
    fields_values = list(form_fields.values())
    if form_fields:
        fieldsets.append(
            {
                "id": "default",
                "title": translate(
                    "label_schema_default",
                    default="Default",
                    domain="plone",
                    context=getRequest(),
                ),
                "fields": fields_values,
                "behavior": "plone",
            }
        )

    # Additional fieldsets (AKA z3c.form groups)
    for group in getattr(form, "groups", []):
        fieldset = {
            "id": group.__name__,
            "title": translate(group.label, context=getRequest()),
            "fields": list(group.fields.values()),
            "behavior": "plone",
        }
        fieldsets.append(fieldset)
    return fieldsets


def get_fieldsets(context, request, schema, additional_schemata=None):
    """Given a base schema, and optionally some additional schemata,
    build a list of fieldsets with the corresponding z3c.form fields in them.
    """
    form = create_form(context, request, schema, additional_schemata)
    return get_form_fieldsets(form)


def get_fieldset_infos(fieldsets):
    """Given a list of fieldset dicts as returned by `get_fieldsets()`,
    return a list of fieldset info dicts that contain the (short) field name
    instead of the actual field instance.
    """
    fieldset_infos = []
    for fieldset in fieldsets:
        fs_info = copy(fieldset)
        fs_info["fields"] = [f.field.getName() for f in fs_info["fields"]]
        fieldset_infos.append(fs_info)
    return fieldset_infos


def get_jsonschema_properties(
    context, request, fieldsets, prefix="", excluded_fields=None
):
    """Build a JSON schema 'properties' list, based on a list of fieldset
    dicts as returned by `get_fieldsets()`.
    """
    properties = OrderedDict()
    if excluded_fields is None:
        excluded_fields = []

    for field in iter_fields(fieldsets):
        fieldname = field.field.getName()
        if fieldname not in excluded_fields:
            # We need to special case relatedItems not to render choices
            # so we try a named adapter first and fallback to unnamed ones.
            adapter = queryMultiAdapter(
                (field.field, context, request),
                interface=IJsonSchemaProvider,
                name=field.__name__,
            )

            adapter = adapter or getMultiAdapter(
                (field.field, context, request), interface=IJsonSchemaProvider
            )

            adapter.prefix = prefix
            if prefix:
                fieldname = ".".join([prefix, fieldname])

            properties[fieldname] = adapter.get_schema()

    return properties


def get_widget_params(schemas):
    params = {}
    for schema in schemas:
        if not schema:
            continue
        tagged_values = mergedTaggedValueDict(schema, WIDGETS_KEY)
        for field_name in schema:
            widget = tagged_values.get(field_name)
            if IParameterizedWidget.providedBy(widget) and widget.params:
                params[field_name] = {}
                for k, v in widget.params.items():
                    if callable(v):
                        v = v()
                    params[field_name][k] = v
    return params


def get_multilingual_directives(schemas):
    params = {}
    for schema in schemas:
        if not schema:
            continue
        tagged_values = mergedTaggedValueList(schema, MULTILINGUAL_KEY)
        result = {field_name: value for _, field_name, value in tagged_values}

        for field_name, value in result.items():
            params[field_name] = {}
            params[field_name]["language_independent"] = value
    return params


def get_jsonschema_for_fti(fti, context, request, excluded_fields=None):
    """Build a complete JSON schema for the given FTI."""
    if excluded_fields is None:
        excluded_fields = []

    # We try..except lookupSchema here, so we still get FTI information
    # through /@types/{typeid} for non-DX type, notably the "Plone Site" type.
    try:
        schema = fti.lookupSchema()
    except AttributeError:
        schema = None
        fieldsets = ()
        additional_schemata = ()
    else:
        additional_schemata = tuple(getAdditionalSchemata(portal_type=fti.id))
        fieldsets = get_fieldsets(context, request, schema, additional_schemata)

    # Build JSON schema properties
    properties = get_jsonschema_properties(
        context, request, fieldsets, excluded_fields=excluded_fields
    )

    required = []
    for field in iter_fields(fieldsets):
        name = field.field.getName()
        # Determine required fields
        if field.field.required:
            required.append(name)

        # Include field modes
        if field.mode:
            properties[name]["mode"] = field.mode

        # Include behavior
        if name in properties:
            behavior = queryUtility(IBehavior, name=field.interface.__identifier__)
            properties[name]["behavior"] = (
                getattr(behavior, "name", None) or field.interface.__identifier__
            )

    return {
        "type": "object",
        "title": translate(fti.Title(), context=getRequest()),
        "properties": IJsonCompatible(properties),
        "required": required,
        "fieldsets": get_fieldset_infos(fieldsets),
        "layouts": getattr(fti, "view_methods", []),
    }


def get_jsonschema_for_portal_type(portal_type, context, request, excluded_fields=None):
    """Build a complete JSON schema for the given portal_type."""
    ttool = getToolByName(context, "portal_types")
    fti = ttool[portal_type]
    return get_jsonschema_for_fti(
        fti, context, request, excluded_fields=excluded_fields
    )


def get_vocab_like_url(endpoint, locator, context, request):
    try:
        context_url = context.absolute_url()
    except AttributeError:
        portal = getSite()
        context_url = portal.absolute_url()
    return "/".join((context_url, endpoint, locator))


def get_vocabulary_url(vocab_name, context, request):
    return get_vocab_like_url("@vocabularies", vocab_name, context, request)


def get_querysource_url(field, context, request):
    return get_vocab_like_url("@querysources", field.getName(), context, request)


def get_source_url(field, context, request):
    return get_vocab_like_url("@sources", field.getName(), context, request)


def serializeSchema(schema):
    """Taken from plone.app.dexterity.serialize
    Finds the FTI and model associated with a schema, and synchronizes
    the schema to the FTI model_source attribute.
    """

    # determine portal_type
    try:
        prefix, portal_type, schemaName = splitSchemaName(schema.__name__)
    except ValueError:
        # not a dexterity schema
        return

    # find the FTI and model
    fti = queryUtility(IDexterityFTI, name=portal_type)
    model = fti.lookupModel()

    # synchronize changes to the model
    syncSchema(schema, model.schemata[schemaName], overwrite=True)
    fti.model_source = serializeModel(model)


def get_info_for_type(context, request, name):
    """Get JSON info for the given portal type"""
    base_context = context

    # If context is not a dexterity content, use site root to get
    # the schema
    if not IDexterityContent.providedBy(context):
        base_context = getSite()

    schema = get_jsonschema_for_portal_type(name, base_context, request)

    if not hasattr(context, "schema"):
        return schema

    # Get the empty fieldsets
    existing = {f.get("id") for f in schema.get("fieldsets", [])}
    generated = set()
    for fieldset in context.schema.queryTaggedValue(FIELDSETS_KEY, []):
        name = fieldset.__name__
        generated.add(name)

        if name not in existing:
            info = get_info_for_fieldset(context, request, name)
            schema["fieldsets"].append(info)
            continue

    # Update fieldset behavior
    for idx, tab in enumerate(schema.get("fieldsets", [])):
        if tab.get("id") in generated:
            schema["fieldsets"][idx]["behavior"] = "plone.dexterity.schema.generated"

    return schema


def get_info_for_field(context, request, name):
    """Get JSON info for the given field name."""
    field = context.publishTraverse(request, name)
    adapter = queryMultiAdapter(
        (field.field, context, request), interface=IJsonSchemaProvider
    )

    schema = adapter.get_schema()
    schema["behavior"] = context.schema.__identifier__
    return IJsonCompatible(schema)


def get_info_for_fieldset(context, request, name):
    """Get JSON info for the given fieldset name."""
    properties = {}
    for fieldset in context.schema.queryTaggedValue(FIELDSETS_KEY, []):
        if name != fieldset.__name__:
            continue

        properties = {
            "id": fieldset.__name__,
            "title": fieldset.label,
            "description": fieldset.description,
            "fields": fieldset.fields,
            "behavior": "plone.dexterity.schema.generated",
        }
    return IJsonCompatible(properties)


def delete_field(context, request, name):
    field = context.publishTraverse(request, name)
    delete = queryMultiAdapter((field, request), name="delete")
    delete()


def delete_fieldset(context, request, name):
    """Taken from plone.schemaeditor 2.x `DeleteFieldset`"""
    new_fieldsets = []
    fieldsets = context.schema.queryTaggedValue(FIELDSETS_KEY, [])
    for fieldset in fieldsets:
        if fieldset.__name__ == name:
            # Can't delete fieldsets with fields
            if fieldset.fields:
                return
            continue
        new_fieldsets.append(fieldset)

    # Nothing changed
    if len(fieldsets) == len(new_fieldsets):
        return

    context.schema.setTaggedValue(FIELDSETS_KEY, new_fieldsets)
    serializeSchema(context.schema)


def add_fieldset(context, request, data):
    name = data.get("id", None)
    title = data.get("title", None)
    description = data.get("description", None)

    if not name:
        name = idnormalizer.normalize(title).replace("-", "_")

    # Default is reserved
    if name == "default":
        return {}

    add = queryMultiAdapter((context, request), name="add-fieldset")
    properties = {"__name__": name, "label": title, "description": description}
    fieldset = add.form_instance.create(data=properties)
    add.form_instance.add(fieldset)

    return get_info_for_fieldset(context, request, name)


def add_field(context, request, data):
    factory = data.get("factory", None)
    title = data.get("title", None)
    description = data.get("description", None)
    required = data.get("required", False)
    name = data.get("id", None)
    if not name:
        name = idnormalizer.normalize(title).replace("-", "_")

    klass = None
    vocabulary = queryUtility(IVocabularyFactory, name="Fields")
    for term in vocabulary(context):
        if factory not in (term.title, term.token):
            continue

        klass = term.value
        break

    if not klass:
        raise BadRequest("Missing/Invalid parameter factory: %s" % factory)

    add = queryMultiAdapter((context, request), name="add-field")
    properties = {
        "title": title,
        "__name__": name,
        "description": description,
        "factory": klass,
        "required": required,
    }

    field = add.form_instance.create(data=properties)
    add.form_instance.add(field)

    return get_info_for_field(context, request, name)


def update_fieldset(context, request, data):
    name = data.get("id", None)
    title = data.get("title", None)
    description = data.get("description", None)
    fields = data.get("fields", None)

    if not name:
        name = idnormalizer.normalize(title).replace("-", "_")

    # We can only re-order fields within the default fieldset
    if name == "default":
        pos = 0
        for field_name in fields:
            if field_name not in context.schema:
                continue

            field = context.publishTraverse(request, field_name)
            order = queryMultiAdapter((field, request), name="order")
            order.move(pos, 0)
            pos += 1
        return

    # Update fieldset
    fieldsets = context.schema.queryTaggedValue(FIELDSETS_KEY, [])
    for idx, fieldset in enumerate(fieldsets):
        if name != fieldset.__name__:
            continue

        if title:
            fieldset.label = title

        if description:
            fieldset.description = description

        pos = 0
        for field_name in fields:
            if field_name not in context.schema:
                continue

            field = context.publishTraverse(request, field_name)
            order = queryMultiAdapter((field, request), name="order")
            order.move(pos, idx + 1)
            pos += 1


def update_field(context, request, data):
    field = context.publishTraverse(request, data.pop("id"))
    edit = queryMultiAdapter((field, request), name="edit")
    default = data.pop("default", _marker)

    properties = {}
    for key, value in data.items():
        key = FIELD_PROPERTIES_MAPPING.get(key, key)
        properties[key] = value

    # clear current min/max to avoid range errors
    if "min" in properties:
        edit.form_instance.field.min = None
    if "max" in properties:
        edit.form_instance.field.max = None

    edit.form_instance.updateFields()
    edit.form_instance.applyChanges(properties)

    if default is not _marker:
        fake_context = FakeDXContext()
        deserializer = queryMultiAdapter(
            (field.field, fake_context, request), IFieldDeserializer
        )
        setattr(field.field, "default", deserializer(default))



from plone.app.layout.navigation.interfaces import INavigationRoot
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.registry.interfaces import IRegistry
from plone.restapi.services.contextnavigation.get import ContextNavigation
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from Products.CMFPlone.tests import dummy
from urllib.parse import urlencode
from zope.component import getUtility
from zope.interface import directlyProvides
from zope.interface import noLongerProvides

import transaction
import unittest


def opts(**kw):
    res = {}
    for k, v in kw.items():
        res["expand.contextnavigation." + k] = v

    return res


class TestServicesContextNavigation(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
    maxDiff = None

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        self.populateSite()
        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def populateSite(self):
        """
        Portal
        +-doc1
        +-doc2
        +-doc3
        +-folder1
          +-doc11
          +-doc12
          +-doc13
        +-link1
        +-folder2
          +-doc21
          +-doc22
          +-doc23
          +-file21
          +-folder21
            +-doc211
            +-doc212
        """
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        if "Members" in self.portal:
            self.portal._delObject("Members")
            self.folder = None
        if "news" in self.portal:
            self.portal._delObject("news")
        if "events" in self.portal:
            self.portal._delObject("events")
        if "front-page" in self.portal:
            self.portal._delObject("front-page")
        if "folder" in self.portal:
            self.portal._delObject("folder")
        if "users" in self.portal:
            self.portal._delObject("users")

        self.portal.invokeFactory("Document", "doc1")
        self.portal.invokeFactory("Document", "doc2")
        self.portal.invokeFactory("Document", "doc3")
        self.portal.invokeFactory("Folder", "folder1")
        self.portal.invokeFactory("Link", "link1")
        self.portal.link1.remoteUrl = "http://plone.org"
        self.portal.link1.reindexObject()
        folder1 = getattr(self.portal, "folder1")
        folder1.invokeFactory("Document", "doc11")
        folder1.invokeFactory("Document", "doc12")
        folder1.invokeFactory("Document", "doc13")
        self.portal.invokeFactory("Folder", "folder2")
        folder2 = getattr(self.portal, "folder2")
        folder2.invokeFactory("Document", "doc21")
        folder2.invokeFactory("Document", "doc22")
        folder2.invokeFactory("Document", "doc23")
        folder2.invokeFactory("File", "file21")
        folder2.invokeFactory("Folder", "folder21")
        folder21 = getattr(folder2, "folder21")
        folder21.invokeFactory("Document", "doc211")
        folder21.invokeFactory("Document", "doc212")

        setRoles(self.portal, TEST_USER_ID, ["Member"])

    def renderer(self, context=None, data=None):
        context = context or self.portal
        request = self.layer["request"]
        request.form.update(data or {})
        return ContextNavigation(context, request)

    def test_contextnavigation_with_no_params_gets_only_top_level(self):
        response = self.api_session.get("/folder1/@contextnavigation")

        self.assertEqual(response.status_code, 200)
        base = self.portal.absolute_url()

        res = {
            "@id": "%s/folder1/@contextnavigation" % base,
            "has_custom_name": False,
            "available": True,
            "items": [
                {
                    "@id": "%s/folder1/doc11" % base,
                    "description": "",
                    "href": "%s/folder1/doc11" % base,
                    "icon": "",
                    "is_current": False,
                    "is_folderish": False,
                    "is_in_path": False,
                    "items": [],
                    "normalized_id": "doc11",
                    "review_state": "private",
                    "thumb": "",
                    "title": "doc11",
                    "type": "document",
                },
                {
                    "@id": "%s/folder1/doc12" % base,
                    "description": "",
                    "href": "%s/folder1/doc12" % base,
                    "icon": "",
                    "is_current": False,
                    "is_folderish": False,
                    "is_in_path": False,
                    "items": [],
                    "normalized_id": "doc12",
                    "review_state": "private",
                    "thumb": "",
                    "title": "doc12",
                    "type": "document",
                },
                {
                    "@id": "%s/folder1/doc13" % base,
                    "description": "",
                    "href": "%s/folder1/doc13" % base,
                    "icon": "",
                    "is_current": False,
                    "is_folderish": False,
                    "is_in_path": False,
                    "items": [],
                    "normalized_id": "doc13",
                    "review_state": "private",
                    "thumb": "",
                    "title": "doc13",
                    "type": "document",
                },
            ],
            "title": "Navigation",
            "url": "%s/sitemap" % base,
        }

        self.assertEqual(
            response.json(),
            res,
        )

    def test_contextnavigation_with_no_params_gets_only_top_level_mixed_content(self):
        # With the context set to folder2 it should return a dict with
        # currentItem set to True
        response = self.api_session.get("/folder2/@contextnavigation")
        self.assertEqual(response.status_code, 200)
        base = self.portal.absolute_url()

        res = {
            "@id": "%s/folder2/@contextnavigation" % base,
            "has_custom_name": False,
            "available": True,
            "items": [
                {
                    "@id": "%s/folder2/doc21" % base,
                    "description": "",
                    "href": "%s/folder2/doc21" % base,
                    "icon": "",
                    "is_current": False,
                    "is_folderish": False,
                    "is_in_path": False,
                    "items": [],
                    "normalized_id": "doc21",
                    "review_state": "private",
                    "thumb": "",
                    "title": "doc21",
                    "type": "document",
                },
                {
                    "@id": "%s/folder2/doc22" % base,
                    "description": "",
                    "href": "%s/folder2/doc22" % base,
                    "icon": "",
                    "is_current": False,
                    "is_folderish": False,
                    "is_in_path": False,
                    "items": [],
                    "normalized_id": "doc22",
                    "review_state": "private",
                    "thumb": "",
                    "title": "doc22",
                    "type": "document",
                },
                {
                    "@id": "%s/folder2/doc23" % base,
                    "description": "",
                    "href": "%s/folder2/doc23" % base,
                    "icon": "",
                    "is_current": False,
                    "is_folderish": False,
                    "is_in_path": False,
                    "items": [],
                    "normalized_id": "doc23",
                    "review_state": "private",
                    "thumb": "",
                    "title": "doc23",
                    "type": "document",
                },
                {
                    "@id": "%s/folder2/file21/view" % base,
                    "description": "",
                    "href": "%s/folder2/file21/view" % base,
                    "icon": None,
                    "is_current": False,
                    "is_folderish": False,
                    "is_in_path": False,
                    "items": [],
                    "normalized_id": "file21",
                    "review_state": "",
                    "thumb": "",
                    "title": "file21",
                    "type": "file",
                },
                {
                    "@id": "%s/folder2/folder21" % base,
                    "description": "",
                    "href": "%s/folder2/folder21" % base,
                    "icon": "",
                    "is_current": False,
                    "is_folderish": True,
                    "is_in_path": False,
                    "items": [],
                    "normalized_id": "folder21",
                    "review_state": "private",
                    "thumb": "",
                    "title": "folder21",
                    "type": "folder",
                },
            ],
            "title": "Navigation",
            "url": "%s/sitemap" % base,
        }
        self.assertEqual(
            response.json(),
            res,
        )

        # self.assertTrue(tree)
        # self.assertEqual(tree["children"][-1]["currentItem"], True)

    def testHeadingLinkRooted(self):
        """
        See that heading link points to a content item if root selected,
        otherwise sitemap.
        """

        q = {
            "expand.contextnavigation.topLevel": 0,
            "expand.contextnavigation.root_path": "/".join(
                self.portal.folder2.getPhysicalPath()[2:]
            ),
        }
        qs = urlencode(q)

        response = self.api_session.get(f"/folder2/@contextnavigation?{qs}")
        self.assertEqual(response.status_code, 200)
        res = response.json()
        base = self.portal.absolute_url()
        self.assertEqual(res["url"], "%s/folder2" % base)

    def testHeadingLinkRootedItemGone(self):
        """
        See that heading link points to a content item which do not exist
        """
        response = self.api_session.get(
            "/folder2/@contextnavigation",
            params={
                "expand.contextnavigation.topLevel": 0,
                "expand.contextnavigation.root_path": "/does/not/exist",
            },
        )
        res = response.json()
        # Points to the site root if the item is gone
        base = self.portal.absolute_url()
        self.assertEqual(res["url"], "%s/sitemap" % base)

    def testHeadingLinkRootless(self):
        """
        See that heading link points to a global sitemap if no root item is set.
        """
        base = self.portal.absolute_url()

        directlyProvides(self.portal.folder2, INavigationRoot)
        transaction.commit()
        response = self.api_session.get(
            "/folder2/@contextnavigation",
            params={"expand.contextnavigation.topLevel": 0},
        )
        link = response.json()["url"]
        # The root is not given -> should render the sitemap in the navigation root
        base = self.portal.absolute_url()
        self.assertEqual(link, "%s/folder2/sitemap" % base)

        # # Even if the assignment contains no topLevel options and no self.root
        # # one should get link to the navigation root sitemap
        # view = self.renderer(
        #     self.portal.folder2.doc21, assignment=navigation.Assignment()
        # )
        response = self.api_session.get(
            "/folder2/doc21/@contextnavigation",
            params={},
        )
        link = response.json()["url"]
        # # The root is not given -> should render the sitemap in the navigation root
        self.assertEqual(link, "%s/folder2/sitemap" % base)

        response = self.api_session.get(
            "/folder1/@contextnavigation",
            params={"expand.contextnavigation.topLevel": 0},
        )
        link = response.json()["url"]
        # The root is not given -> should render the sitemap in the navigation root
        self.assertEqual(link, "%s/sitemap" % base)

        noLongerProvides(self.portal.folder2, INavigationRoot)
        transaction.commit()

    def testNavTreeExcludesItemsWithExcludeProperty(self):
        # Make sure that items with the exclude_from_nav property set get
        # no_display set to True
        base = self.portal.absolute_url()

        self.portal.folder2.exclude_from_nav = True
        self.portal.folder2.reindexObject()

        transaction.commit()

        response = self.api_session.get(
            "@contextnavigation",
            params={
                "expand.contextnavigation.includeTop": True,
                "expand.contextnavigation.topLevel": 0,
                "expand.contextnavigation.bottomLevel": 0,
            },
        )
        tree = response.json()

        for c in tree["items"]:
            if c["href"] == "%s/folder2" % base:
                self.fail()

        self.portal.folder2.exclude_from_nav = False
        self.portal.folder2.reindexObject()
        transaction.commit()

    def testNavTreeExcludesDefaultPage(self):
        # Make sure that items which are the default page are excluded
        base = self.portal.absolute_url()
        response = self.api_session.get(
            "/folder2/@contextnavigation",
            params={},
        )
        tree = response.json()
        self.assertTrue(
            [
                item
                for item in tree["items"]
                if item["href"] == "%s/folder2/doc21" % base
            ]
        )

        self.portal.folder2.setDefaultPage("doc21")
        transaction.commit()

        response = self.api_session.get(
            "/folder2/@contextnavigation",
            params={},
        )
        tree = response.json()
        self.assertFalse(
            [
                item
                for item in tree["items"]
                if item["href"] == "%s/folder2/doc21" % base
            ]
        )

        self.portal.folder2.setDefaultPage(None)
        transaction.commit()

    def testPortletsTitle(self):
        """If portlet's name is not explicitely specified we show
        default fallback 'Navigation', translate it and hide it
        with CSS."""
        response = self.api_session.get(
            "/@contextnavigation",
            params={},
        )
        tree = response.json()
        self.assertEqual(tree["title"], "Navigation")

        response = self.api_session.get(
            "/@contextnavigation",
            params={"expand.contextnavigation.name": "New navigation title"},
        )
        tree = response.json()
        self.assertEqual(tree["title"], "New navigation title")

    def testTopLevelTooDeep(self):

        view = self.renderer(self.portal, opts(topLevel=5))
        tree = view(expand=True)

        self.assertEqual(len(tree["contextnavigation"]["items"]), 0)

    def testShowAllParentsOverridesNavTreeExcludesItemsWithExcludeProperty(self):
        # Make sure that items whose ids are in the idsNotToList navTree
        # property are not included
        self.portal.folder2.exclude_from_nav = True
        self.portal.folder2.reindexObject()
        view = self.renderer(
            self.portal.folder2.doc21, opts(includeTop=True, topLevel=0)
        )
        tree = view(expand=True)

        found = False
        base = self.portal.absolute_url()

        for c in tree["contextnavigation"]["items"]:
            if c["href"] == "%s/folder2" % base:
                found = True
                break

        self.assertTrue(found)

    # # this test is not needed, we don't expose show_children
    # def testNavTreeMarksParentMetaTypesNotToQuery(self):
    #     # Make sure that items whose ids are in the idsNotToList navTree
    #     # property get no_display set to True
    #     view = self.renderer(self.portal.folder2.file21)
    #     tree = view(expand=True)
    #
    #     self.assertEqual(tree["contextnavigation"]["items"][-1]["show_children"], True)
    #
    #     registry = self.portal.portal_registry
    #     registry["plone.parent_types_not_to_query"] = [u"Folder"]
    #
    #     view = self.renderer(self.portal.folder2.file21)
    #     tree = view(expand=True)
    #
    #     self.assertEqual(tree["contextnavigation"]["items"][-1]["show_children"], False)

    def testCreateNavTreeWithLink(self):
        view = self.renderer(self.portal)
        tree = view(expand=True)["contextnavigation"]

        for child in tree["items"]:
            if child["portal_type"] != "Link":
                self.assertFalse(child["getRemoteUrl"])

            if child["Title"] == "link1":
                self.assertEqual(child["getRemoteUrl"], "http://plone.org")
                # as Creator, link1 should not use the remote Url
                self.assertFalse(child["useRemoteUrl"])

        self.portal.link1.setCreators(["some_other_user"])
        self.portal.link1.reindexObject()
        view = self.renderer(self.portal)
        tree = view(expand=True)["contextnavigation"]

        for child in tree["items"]:
            if child["portal_type"] != "Link":
                self.assertFalse(child["getRemoteUrl"])
            if child["Title"] == "link1":
                self.assertEqual(child["getRemoteUrl"], "http://plone.org")
                # as non-Creator user, link1 should use the remote Url
                self.assertTrue(child["useRemoteUrl"])

    def testNonStructuralFolderHidesChildren(self):
        base = self.portal.absolute_url()
        # Make sure NonStructuralFolders act as if parent_types_not_to_query
        # is set.
        f = dummy.NonStructuralFolder("ns_folder")
        self.portal.folder1._setObject("ns_folder", f)
        self.portal.portal_catalog.reindexObject(self.portal.folder1.ns_folder)
        self.portal.portal_catalog.reindexObject(self.portal.folder1)
        view = self.renderer(
            self.portal.folder1.ns_folder, opts(includeTop=True, topLevel=0)
        )
        tree = view.getNavTree()
        self.assertEqual(
            tree["items"][3]["items"][3]["href"],
            "%s/folder1/ns_folder" % base,
        )
        self.assertEqual(len(tree["items"][3]["items"][3]["items"]), 0)

    def testTopLevel(self):
        base = self.portal.absolute_url()
        view = self.renderer(self.portal.folder2.file21, opts(topLevel=1))
        tree = view.getNavTree()
        self.assertTrue(tree)

        self.assertEqual(
            tree["items"][-1]["href"],
            "%s/folder2/folder21" % base,
        )

    def testTopLevelWithContextAboveLevel(self):
        view = self.renderer(self.portal, opts(topLevel=1))
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertEqual(len(tree["items"]), 0)

    def testIncludeTopWithoutNavigationRoot(self):
        base = self.portal.absolute_url()
        view = self.renderer(
            self.portal.folder2.folder21,
            opts(topLevel=0, root_path=None, includeTop=True),
        )
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertEqual(len(tree["items"]), 6)

        # self.assertTrue(view.root_is_portal())
        self.assertEqual(tree["url"], "%s/sitemap" % base)

    def testTopLevelWithNavigationRoot(self):
        base = self.portal.absolute_url()
        # self.portal.folder2.invokeFactory("Folder", "folder21")
        # self.portal.folder2.folder21.invokeFactory("Document", "doc211")
        view = self.renderer(
            self.portal.folder2.folder21,
            opts(
                topLevel=1,
                root_path="/folder2"
                # self.portal.folder2.UID()
            ),
        )
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertEqual(len(tree["items"]), 2)
        self.assertEqual(
            tree["items"][0]["href"],
            "%s/folder2/folder21/doc211" % base,
        )

    def testMultipleTopLevelWithNavigationRoot(self):
        # See bug 9405
        # http://dev.plone.org/plone/ticket/9405
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        self.portal.invokeFactory("Folder", "abc")
        self.portal.invokeFactory("Folder", "abcde")
        self.portal.abc.invokeFactory("Folder", "down_abc")
        self.portal.abcde.invokeFactory("Folder", "down_abcde")

        view1 = self.renderer(
            self.portal.abc,
            opts(topLevel=0, root_path="/abc"),
        )
        view2 = self.renderer(
            self.portal.abc,
            opts(topLevel=0, root_path="/abcde"),
        )

        tree1 = view1.getNavTree()
        tree2 = view2.getNavTree()
        self.assertEqual(len(tree1["items"]), 1)
        self.assertEqual(len(tree2["items"]), 1)

        view1 = self.renderer(
            self.portal.abcde,
            opts(topLevel=0, root_path="/abc"),
        )

        view2 = self.renderer(
            self.portal.abcde,
            opts(topLevel=0, root_path="/abcde"),
        )

        tree1 = view1.getNavTree()
        tree2 = view2.getNavTree()

        self.assertEqual(len(tree2["items"]), 1)
        self.assertEqual(len(tree1["items"]), 1)

    def testShowAllParentsOverridesBottomLevel(self):
        base = self.portal.absolute_url()
        view = self.renderer(
            self.portal.folder2.file21,
            opts(bottomLevel=1, topLevel=0),
        )
        tree = view.getNavTree()
        self.assertTrue(tree)
        # Note: showAllParents makes sure we actually return items on the,
        # path to the context, but the portlet will not display anything
        # below bottomLevel.
        self.assertEqual(tree["items"][-1]["href"], "%s/folder2" % base)
        # self.assertEqual(len(tree["items"][-1]["items"]), 1)
        # self.assertEqual(
        #     tree["items"][-1]["items"][0]["href"],
        #     "http://localhost:55001/plone/folder2/file21",
        # )

    def testBottomLevelStopsAtFolder(self):
        base = self.portal.absolute_url()
        view = self.renderer(
            self.portal.folder2,
            opts(bottomLevel=1, topLevel=0),
        )
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertEqual(tree["items"][-1]["href"], "%s/folder2" % base)
        self.assertEqual(len(tree["items"][-1]["items"]), 0)

    def testBottomLevelZeroNoLimit(self):
        """Test that bottomLevel=0 means no limit for bottomLevel."""
        base = self.portal.absolute_url()

        # first we set a high integer as bottomLevel to simulate "no limit"
        view = self.renderer(
            self.portal.folder2,
            opts(bottomLevel=99, topLevel=0),
        )
        tree = view.getNavTree()
        self.assertTrue(tree)

        self.assertEqual(
            tree["items"][-1]["items"][0]["href"],
            "%s/folder2/doc21" % base,
        )

        # now set bottomLevel to 0 -> outcome should be the same
        view = self.renderer(
            self.portal.folder2,
            opts(bottomLevel=0, topLevel=0),
        )
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertEqual(
            tree["items"][-1]["items"][0]["href"],
            "%s/folder2/doc21" % base,
        )

    def testBottomLevelZeroNoLimitRendering(self):
        """Test that bottomLevel=0 means no limit for bottomLevel."""

        # first we set a high integer as bottomLevel to simulate "no limit"
        view = self.renderer(
            self.portal.folder2,
            opts(bottomLevel=99, topLevel=0),
        )
        a = view(expand=True)

        # now set bottomLevel to 0 -> outcome should be the same
        view = self.renderer(
            self.portal.folder2,
            opts(bottomLevel=0, topLevel=0),
        )
        b = view(expand=True)

        self.assertEqual(a, b)

    def testNavRootWithUnicodeNavigationRoot(self):
        # self.portal.folder2.invokeFactory("Folder", "folder21")
        # self.portal.folder2.folder21.invokeFactory("Document", "doc211")
        base = self.portal.absolute_url()
        view = self.renderer(
            self.portal.folder2.folder21,
            opts(
                topLevel=1,
                root_path="/folder2",
            ),
        )
        tree = view.getNavTree()
        self.assertEqual(tree["url"], "%s/folder2/folder21" % base)

    def testNoRootSet(self):
        base = self.portal.absolute_url()
        view = self.renderer(
            self.portal.folder2.file21,
            opts(root_path="", topLevel=0),
        )
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertEqual(tree["items"][-1]["href"], "%s/folder2" % base)

    def testRootIsNotPortal(self):
        base = self.portal.absolute_url()
        view = self.renderer(
            self.portal.folder2.file21,
            opts(root_path="/folder2", topLevel=0),
        )
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertEqual(tree["items"][0]["href"], "%s/folder2/doc21" % base)

    def testRootDoesNotExist(self):
        view = self.renderer(
            self.portal.folder2.file21,
            opts(root_path="DOESNT_EXIST", topLevel=0),
        )
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertEqual(len(tree["items"]), 6)

    def testAboveRoot(self):
        base = self.portal.absolute_url()
        registry = getUtility(IRegistry)
        registry["plone.root"] = "/folder2"
        view = self.renderer(self.portal, opts(topLevel=0))
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertEqual(tree["items"][0]["href"], "%s/folder2/doc21" % base)

    def testOutsideRoot(self):
        base = self.portal.absolute_url()
        view = self.renderer(
            self.portal.folder1,
            opts(root_path="/folder2", topLevel=0),
        )
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertEqual(tree["items"][0]["href"], "%s/folder2/doc21" % base)

    def testRootIsCurrent(self):
        base = self.portal.absolute_url()
        view = self.renderer(
            self.portal.folder2,
            opts(currentFolderOnly=True),
        )
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertEqual(tree["items"][0]["href"], "%s/folder2/doc21" % base)

    def testRootIsCurrentWithFolderishDefaultPage(self):
        # self.portal.folder2.invokeFactory("Folder", "folder21")
        base = self.portal.absolute_url()
        self.portal.folder2.setDefaultPage("folder21")

        view = self.renderer(
            self.portal.folder2.folder21,
            opts(currentFolderOnly=True),
        )
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertEqual(tree["items"][0]["href"], "%s/folder2/doc21" % base)

    def testCustomQuery(self):
        # Try a custom query script for the navtree that returns only published
        # objects
        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        workflow = self.portal.portal_workflow
        factory = self.portal.manage_addProduct["PythonScripts"]
        factory.manage_addPythonScript("getCustomNavQuery")
        script = self.portal.getCustomNavQuery
        script.ZPythonScript_edit("", 'return {"review_state": "published"}')

        self.assertEqual(self.portal.getCustomNavQuery(), {"review_state": "published"})

        view = self.renderer(self.portal.folder2, opts(topLevel=0))
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertTrue("items" in tree)

        # Should only contain current object
        self.assertEqual(len(tree["items"]), 1)  # different

        # change workflow for folder1
        workflow.doActionFor(self.portal.folder1, "publish")
        self.portal.folder1.reindexObject()

        view = self.renderer(self.portal.folder2, opts(topLevel=0))
        tree = view.getNavTree()
        # Should only contain current object and published folder
        self.assertEqual(len(tree["items"]), 2)

    def testStateFiltering(self):
        # Test Navtree workflow state filtering
        from Products.CMFPlone.interfaces import INavigationSchema  # noqa

        setRoles(self.portal, TEST_USER_ID, ["Manager"])
        registry = getUtility(IRegistry)
        navigation_settings = registry.forInterface(INavigationSchema, prefix="plone")
        workflow = self.portal.portal_workflow
        navigation_settings.workflow_states_to_show = ("published",)
        navigation_settings.filter_on_workflow = True

        view = self.renderer(self.portal.folder2, opts(topLevel=0))
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertTrue("items" in tree)

        # Should only contain current object
        self.assertEqual(len(tree["items"]), 1)

        # change workflow for folder1
        workflow.doActionFor(self.portal.folder1, "publish")
        self.portal.folder1.reindexObject()
        view = self.renderer(self.portal.folder2, opts(topLevel=0))
        tree = view.getNavTree()

        # Should only contain current object and published folder
        self.assertEqual(len(tree["items"]), 2)

    def testPrunedRootNode(self):
        # This test has been changed to conform to reality
        registry = self.portal.portal_registry
        registry["plone.parent_types_not_to_query"] = ["Folder"]

        view = self.renderer(self.portal.folder1, opts(topLevel=0))
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertEqual(len(tree["items"][4]["items"]), 0)

    def testPrunedRootNodeShowsAllParents(self):
        registry = self.portal.portal_registry
        registry["plone.parent_types_not_to_query"] = ["Folder"]

        view = self.renderer(self.portal.folder1.doc11, opts(topLevel=1))
        tree = view.getNavTree()
        self.assertTrue(tree)
        self.assertEqual(len(tree["items"]), 1)
        self.assertTrue(tree["items"][0]["href"].endswith("/plone/folder1/doc11"))

    def testIsCurrentParentWithOverlapingNames(self):
        setRoles(
            self.portal,
            TEST_USER_ID,
            [
                "Manager",
            ],
        )
        self.portal.invokeFactory("Folder", "folder2x")
        self.portal.folder2x.invokeFactory("Document", "doc2x1")
        setRoles(
            self.portal,
            TEST_USER_ID,
            [
                "Member",
            ],
        )
        view = self.renderer(self.portal.folder2x.doc2x1, opts(topLevel=0))
        tree = view.getNavTree()
        self.assertTrue(tree)

        folder2x_node = [
            n for n in tree["items"] if n["href"].endswith("/plone/folder2x")
        ][0]
        self.assertTrue(folder2x_node["is_in_path"])

        folder2_node = [
            n for n in tree["items"] if n["href"].endswith("/plone/folder2")
        ][0]
        self.assertFalse(folder2_node["is_in_path"])

    def testPortletNotDisplayedOnINavigationRoot(self):
        """test that navigation portlet does not show on INavigationRoot
        folder
        """
        self.assertFalse(INavigationRoot.providedBy(self.portal.folder1))

        # make folder1 as navigation root
        directlyProvides(self.portal.folder1, INavigationRoot)
        self.assertTrue(INavigationRoot.providedBy(self.portal.folder1))

        # add nested subfolder in folder1
        self.portal.folder1.invokeFactory("Folder", "folder1_1")

        # make a navigation portlet
        view = self.renderer(self.portal.folder1, opts(bottomLevel=0, topLevel=1))
        tree = view(expand=True)
        self.assertTrue(tree)
        # check there is no portlet
        self.assertFalse(tree["contextnavigation"]["items"])

    def testINavigationRootWithRelativeRootSet(self):
        """test that navigation portlet uses relative root set by user
        even in INavigationRoot case.
        """
        self.assertFalse(INavigationRoot.providedBy(self.portal.folder1))

        # make folder1 as navigation root
        directlyProvides(self.portal.folder1, INavigationRoot)
        self.assertTrue(INavigationRoot.providedBy(self.portal.folder1))

        # add two nested subfolders in folder1
        self.portal.folder1.invokeFactory("Folder", "folder1_1")
        self.portal.folder1.folder1_1.invokeFactory("Folder", "folder1_1_1")

        # make a navigation portlet with navigation root set
        view = self.renderer(
            self.portal.folder1.folder1_1,
            opts(bottomLevel=0, topLevel=0, root_path="/folder1/folder1_1"),
        )
        tree = view(expand=True)["contextnavigation"]

        # check there is a portlet
        self.assertTrue(tree["items"])

        # check that portlet root is actually the one specified
        self.assertTrue(tree["url"].endswith("/plone/folder1/folder1_1"))

        # check that portlet tree actually includes children
        self.assertEqual(len(tree["items"]), 1)
        self.assertTrue(
            tree["items"][0]["href"].endswith(
                "/plone/folder1/folder1_1/folder1_1_1",
            )
        )

    def testServiceId(self):
        view = self.renderer(
            self.portal.folder2.file21,
            opts(root_path="", topLevel=0),
        )
        portlet = view(expand=True)

        self.assertTrue(
            portlet["contextnavigation"]["@id"].endswith(
                "/plone/folder2/file21/@contextnavigation",
            )
        )
        portlet = view(expand=False)
        self.assertEqual(len(portlet["contextnavigation"]), 1)

    def testContextNavigation(self):
        response = self.api_session.get("/folder1?expand=contextnavigation")
        res = response.json()
        self.assertTrue(
            res["@components"]["contextnavigation"]["items"][0]["@id"].endswith(
                "/plone/folder1/doc11",
            )
        )



from AccessControl import getSecurityManager
from importlib import import_module
from plone.autoform.interfaces import READ_PERMISSIONS_KEY
from plone.dexterity.utils import iterSchemata
from plone.restapi.batching import HypermediaBatch
from plone.restapi.interfaces import IBlockFieldSerializationTransformer
from plone.restapi.interfaces import IFieldSerializer
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import ISerializeToJsonSummary
from plone.restapi.serializer.converters import json_compatible
from plone.restapi.serializer.expansion import expandable_elements
from plone.restapi.services.locking import lock_info
from plone.supermodel.utils import mergedTaggedValueDict
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces import IPloneSiteRoot
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.component import queryMultiAdapter
from zope.component import queryUtility
from zope.component import subscribers
from zope.interface import implementer
from zope.interface import Interface
from zope.schema import getFields
from zope.security.interfaces import IPermission

import json

HAS_PLONE_6 = getattr(
    import_module("Products.CMFPlone.factory"), "PLONE60MARKER", False
)


@implementer(ISerializeToJson)
@adapter(IPloneSiteRoot, Interface)
class SerializeSiteRootToJson:
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def _build_query(self):
        path = "/".join(self.context.getPhysicalPath())
        query = {
            "path": {"depth": 1, "query": path},
            "sort_on": "getObjPositionInParent",
        }
        return query

    def __call__(self, version=None):
        version = "current" if version is None else version
        if version != "current":
            return {}

        query = self._build_query()

        catalog = getToolByName(self.context, "portal_catalog")
        brains = catalog(query)

        batch = HypermediaBatch(self.request, brains)

        result = {
            # '@context': 'http://www.w3.org/ns/hydra/context.jsonld',
            "@id": self.context.absolute_url(),
            "id": self.context.id,
            "@type": "Plone Site",
            "title": self.context.Title(),
            "parent": {},
            "is_folderish": True,
            "description": self.context.description,
        }

        if HAS_PLONE_6:
            # Insert Plone Site DX root field values
            for schema in iterSchemata(self.context):
                read_permissions = mergedTaggedValueDict(schema, READ_PERMISSIONS_KEY)

                for name, field in getFields(schema).items():

                    if not self.check_permission(
                        read_permissions.get(name), self.context
                    ):
                        continue

                    # serialize the field
                    serializer = queryMultiAdapter(
                        (field, self.context, self.request), IFieldSerializer
                    )
                    value = serializer()
                    result[json_compatible(name)] = value

            # Insert locking information
            result.update({"lock": lock_info(self.context)})
        else:
            # Apply the fake blocks behavior in site root hack using site root properties
            result.update(
                {
                    "blocks": self.serialize_blocks(),
                    "blocks_layout": json.loads(
                        getattr(self.context, "blocks_layout", "{}")
                    ),
                }
            )

        # Insert expandable elements
        result.update(expandable_elements(self.context, self.request))

        result["items_total"] = batch.items_total
        if batch.links:
            result["batching"] = batch.links

        result["items"] = [
            getMultiAdapter((brain, self.request), ISerializeToJsonSummary)()
            for brain in batch
        ]

        return result

    def check_permission(self, permission_name, obj):
        if permission_name is None:
            return True

        if permission_name not in self.permission_cache:
            permission = queryUtility(IPermission, name=permission_name)
            if permission is None:
                self.permission_cache[permission_name] = True
            else:
                sm = getSecurityManager()
                self.permission_cache[permission_name] = bool(
                    sm.checkPermission(permission.title, obj)
                )
        return self.permission_cache[permission_name]

    def serialize_blocks(self):
        # This is only for below 6
        blocks = json.loads(getattr(self.context, "blocks", "{}"))
        if not blocks:
            return blocks
        for id, block_value in blocks.items():
            block_type = block_value.get("@type", "")
            handlers = []
            for h in subscribers(
                (self.context, self.request), IBlockFieldSerializationTransformer
            ):
                if h.block_type == block_type or h.block_type is None:
                    handlers.append(h)

            for handler in sorted(handlers, key=lambda h: h.order):
                if not getattr(handler, "disabled", False):
                    block_value = handler(block_value)

            blocks[id] = block_value
        return blocks



from plone import api
from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession

import transaction
import unittest


class TestPrincipalsEndpoint(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        properties = {
            "email": "noam.chomsky@example.com",
            "username": "noamchomsky",
            "fullname": "Noam Avram Chomsky",
            "home_page": "web.mit.edu/chomsky",
            "description": "Professor of Linguistics",
            "location": "Cambridge, MA",
        }
        self.user = api.user.create(
            email="noam.chomsky@example.com", username="noam", properties=properties
        )

        self.gtool = api.portal.get_tool("portal_groups")
        properties = {
            "title": "Plone Team",
            "description": "We are Plone",
            "email": "ploneteam@plone.org",
        }
        self.gtool.addGroup(
            "ploneteam",
            (),
            (),
            properties=properties,
            title=properties["title"],
            description=properties["description"],
        )
        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_get_principals(self):
        response = self.api_session.get("/@principals", params={"search": "noam"})
        self.assertEqual(200, response.status_code)

        response = response.json()
        self.assertEqual(2, len(response))
        self.assertEqual(1, len(response["users"]))
        self.assertEqual("noam", response["users"][0]["id"])

        response = self.api_session.get("/@principals", params={"search": "plone"})
        self.assertEqual(200, response.status_code)

        response = response.json()
        self.assertEqual(2, len(response))
        self.assertEqual(1, len(response["groups"]))
        self.assertEqual("ploneteam", response["groups"][0]["id"])

    def test_get_principals_response_both(self):
        self.user = api.user.create(
            email="plone.user@example.com", username="plone.user"
        )
        transaction.commit()

        response = self.api_session.get("/@principals", params={"search": "plone"})
        self.assertEqual(200, response.status_code)

        response = response.json()
        self.assertEqual(2, len(response))
        self.assertEqual(1, len(response["users"]))
        self.assertEqual(1, len(response["groups"]))
        self.assertEqual("plone.user", response["users"][0]["id"])
        self.assertEqual("ploneteam", response["groups"][0]["id"])



from datetime import date
from decimal import Decimal
from plone.app.multilingual.dx import directives
from plone.app.textfield import RichText
from plone.autoform import directives as form
from plone.dexterity.fti import DexterityFTI
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from plone.restapi.types.interfaces import IJsonSchemaProvider
from plone.restapi.types.utils import get_fieldsets
from plone.restapi.types.utils import get_jsonschema_for_fti
from plone.restapi.types.utils import get_jsonschema_for_portal_type
from plone.restapi.types.utils import get_jsonschema_properties
from plone.schema import Email
from plone.schema import JSONField
from plone.supermodel import model
from Products.CMFCore.utils import getToolByName
from unittest import TestCase
from z3c.form.browser.text import TextWidget
from zope import schema
from zope.component import getMultiAdapter
from zope.interface import provider
from zope.schema.interfaces import IContextAwareDefaultFactory
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary


class IDummySchema(model.Schema):

    field1 = schema.Bool(title="Foo", description="")

    field2 = schema.TextLine(title="Bar", description="")


class ITaggedValuesSchema(model.Schema):

    form.mode(field_mode_hidden="hidden")
    field_mode_hidden = schema.TextLine(title="ModeHidden", description="")

    form.mode(field_mode_display="display")
    field_mode_display = schema.TextLine(title="ModeDisplay", description="")

    form.mode(field_mode_input="input")
    field_mode_input = schema.TextLine(title="ModeInput", description="")

    field_mode_default = schema.TextLine(title="ModeInput", description="")

    parametrized_widget_field = schema.TextLine(title="Parametrized widget field")
    form.widget(
        "parametrized_widget_field", a_param="some_value", defaultFactory=lambda: "Foo"
    )

    not_parametrized_widget_field = schema.TextLine(
        title="No parametrized widget field"
    )
    form.widget(not_parametrized_widget_field=TextWidget)

    directives.languageindependent("test_language_independent_field")
    test_language_independent_field = schema.TextLine(
        required=False,
    )


class TestJsonSchemaUtils(TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

    def test_get_jsonschema_properties(self):
        fieldsets = get_fieldsets(self.portal, self.request, IDummySchema)
        info = get_jsonschema_properties(self.portal, self.request, fieldsets)
        expected = {
            "field1": {
                "title": "Foo",
                "description": "",
                "factory": "Yes/No",
                "type": "boolean",
            },
            "field2": {
                "title": "Bar",
                "description": "",
                "factory": "Text line (String)",
                "type": "string",
            },
        }
        self.assertEqual(info, expected)

    def test_get_jsonschema_for_fti(self):
        portal = self.portal
        request = self.request
        ttool = getToolByName(portal, "portal_types")
        jsonschema = get_jsonschema_for_fti(ttool["Document"], portal, request)
        self.assertEqual(jsonschema["title"], "Page")
        self.assertEqual(jsonschema["type"], "object")
        self.assertIn("title", list(jsonschema["properties"]))
        self.assertIn("title", jsonschema["required"])
        self.assertEqual("default", jsonschema["fieldsets"][0]["id"])
        self.assertIn("title", jsonschema["fieldsets"][0]["fields"])
        self.assertIn("layouts", jsonschema)

        jsonschema = get_jsonschema_for_fti(
            ttool["Document"], portal, request, excluded_fields=["title"]
        )
        self.assertNotIn("title", list(jsonschema["properties"]))

    def test_get_jsonschema_for_fti_non_dx(self):
        """Make sure FTIs without lookupSchema are supported."""
        fti = self.portal.portal_types["Discussion Item"]
        self.assertFalse(hasattr(fti, "lookupSchema"))

        # This shouldn't raise an error.
        get_jsonschema_for_fti(fti, self.portal, self.request)

    def test_get_jsonschema_for_portal_type(self):
        portal = self.portal
        request = self.request
        jsonschema = get_jsonschema_for_portal_type("Document", portal, request)
        self.assertEqual(jsonschema["title"], "Page")
        self.assertEqual(jsonschema["type"], "object")
        self.assertIn("title", list(jsonschema["properties"]))
        self.assertIn("title", jsonschema["required"])
        self.assertEqual("default", jsonschema["fieldsets"][0]["id"])
        self.assertIn("title", jsonschema["fieldsets"][0]["fields"])

        jsonschema = get_jsonschema_for_portal_type(
            "Document", portal, request, excluded_fields=["title"]
        )
        self.assertNotIn("title", list(jsonschema["properties"]))


class TestTaggedValuesJsonSchemaUtils(TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        fti = DexterityFTI("TaggedDocument")
        self.portal.portal_types._setObject("TaggedDocument", fti)
        fti.klass = "plone.dexterity.content.Container"
        fti.schema = "plone.restapi.tests.test_types.ITaggedValuesSchema"

    def test_get_jsonschema_with_hidden_field(self):
        ttool = getToolByName(self.portal, "portal_types")
        jsonschema = get_jsonschema_for_fti(
            ttool["TaggedDocument"], self.portal, self.request
        )

        self.assertEqual(
            "hidden", jsonschema["properties"]["field_mode_hidden"]["mode"]
        )
        self.assertEqual(
            "display", jsonschema["properties"]["field_mode_display"]["mode"]
        )
        self.assertEqual("input", jsonschema["properties"]["field_mode_input"]["mode"])

    def test_get_jsonschema_with_widget_params(self):
        ttool = getToolByName(self.portal, "portal_types")
        jsonschema = get_jsonschema_for_fti(
            ttool["TaggedDocument"], self.portal, self.request
        )
        self.assertEqual(
            "some_value",
            jsonschema["properties"]["parametrized_widget_field"]["widgetOptions"][
                "a_param"
            ],
        )

    def test_do_not_fail_with_non_parametrized_widget(self):
        ttool = getToolByName(self.portal, "portal_types")
        jsonschema = get_jsonschema_for_fti(
            ttool["TaggedDocument"], self.portal, self.request
        )
        self.assertEqual(
            "No parametrized widget field",
            jsonschema["properties"]["not_parametrized_widget_field"]["title"],
        )

    def test_resolve_callable_widget_params(self):
        ttool = getToolByName(self.portal, "portal_types")
        jsonschema = get_jsonschema_for_fti(
            ttool["TaggedDocument"], self.portal, self.request
        )

        self.assertEqual(
            "Foo",
            jsonschema["properties"]["parametrized_widget_field"]["widgetOptions"].get(
                "defaultFactory"
            ),
        )

    def test_get_jsonschema_with_language_independent_fields(self):
        ttool = getToolByName(self.portal, "portal_types")
        jsonschema = get_jsonschema_for_fti(
            ttool["TaggedDocument"], self.portal, self.request
        )

        self.assertIn(
            "multilingual_options",
            jsonschema["properties"]["test_language_independent_field"],
        )
        self.assertIn(
            "language_independent",
            jsonschema["properties"]["test_language_independent_field"][
                "multilingual_options"
            ],
        )
        self.assertTrue(
            jsonschema["properties"]["test_language_independent_field"][
                "multilingual_options"
            ]
        )


class TestJsonSchemaProviders(TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self.dummy_vocabulary = SimpleVocabulary(
            [
                SimpleTerm(value="foo", title="Foo"),
                SimpleTerm(value="bar", title="Bar"),
            ]
        )

    from zope.interface import provider
    from zope.schema.interfaces import IContextSourceBinder

    @provider(IContextSourceBinder)
    def dummy_source_vocab(self, context):
        return SimpleVocabulary(
            [
                SimpleTerm(value="foo", title="Foo"),
                SimpleTerm(value="bar", title="Bar"),
            ]
        )

    def test_context_aware_default_factory(self):
        folder = self.portal[
            self.portal.invokeFactory("Folder", id="folder", title="My Folder")
        ]

        @provider(IContextAwareDefaultFactory)
        def uppercased_title_default(context):
            return context.title.upper()

        field = schema.TextLine(
            title="My field",
            description="My great field",
            defaultFactory=uppercased_title_default,
        )
        adapter = getMultiAdapter((field, folder, self.request), IJsonSchemaProvider)

        self.assertEqual(
            {
                "type": "string",
                "title": "My field",
                "factory": "Text line (String)",
                "description": "My great field",
                "default": "MY FOLDER",
            },
            adapter.get_schema(),
        )

    def test_textline(self):
        field = schema.TextLine(
            title="My field", description="My great field", default="foobar"
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "string",
                "title": "My field",
                "factory": "Text line (String)",
                "description": "My great field",
                "default": "foobar",
            },
            adapter.get_schema(),
        )

    def test_text(self):
        field = schema.Text(
            title="My field",
            description="My great field",
            default="Lorem ipsum dolor sit amet",
            min_length=10,
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "string",
                "title": "My field",
                "description": "My great field",
                "factory": "Text",
                "widget": "textarea",
                "default": "Lorem ipsum dolor sit amet",
                "minLength": 10,
            },
            adapter.get_schema(),
        )

    def test_bool(self):
        field = schema.Bool(
            title="My field", description="My great field", default=False
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "boolean",
                "title": "My field",
                "description": "My great field",
                "factory": "Yes/No",
                "default": False,
            },
            adapter.get_schema(),
        )

    def test_float(self):
        field = schema.Float(
            title="My field",
            description="My great field",
            min=0.0,
            max=1.0,
            default=0.5,
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "minimum": 0.0,
                "maximum": 1.0,
                "type": "number",
                "title": "My field",
                "description": "My great field",
                "factory": "Floating-point number",
                "default": 0.5,
            },
            adapter.get_schema(),
        )

    def test_email(self):
        field = Email(
            title="Email",
            description="Email field",
            default="foo@bar.com",
            min_length=10,
            max_length=20,
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "string",
                "title": "Email",
                "description": "Email field",
                "factory": "Email",
                "widget": "email",
                "default": "foo@bar.com",
                "minLength": 10,
                "maxLength": 20,
            },
            adapter.get_schema(),
        )

    def test_password(self):
        field = schema.Password(
            title="Password",
            description="Password field",
            default="secret",
            min_length=4,
            max_length=8,
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "string",
                "title": "Password",
                "description": "Password field",
                "factory": "Password",
                "widget": "password",
                "default": "secret",
                "minLength": 4,
                "maxLength": 8,
            },
            adapter.get_schema(),
        )

    def test_uri(self):
        field = schema.URI(
            title="URI",
            description="URI field",
            default="http://foo.bar",
            min_length=10,
            max_length=100,
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "string",
                "title": "URI",
                "description": "URI field",
                "factory": "URL",
                "widget": "url",
                "default": "http://foo.bar",
                "minLength": 10,
                "maxLength": 100,
            },
            adapter.get_schema(),
        )

    def test_decimal(self):
        field = schema.Decimal(
            title="My field",
            description="My great field",
            min=Decimal(0),
            max=Decimal(1),
            default=Decimal(0.5),
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "minimum": 0.0,
                "maximum": 1.0,
                "type": "number",
                "factory": "Floating-point number",
                "title": "My field",
                "description": "My great field",
                "default": 0.5,
            },
            adapter.get_schema(),
        )

    def test_int(self):
        field = schema.Int(
            title="My field", description="My great field", min=0, max=100, default=50
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "minimum": 0,
                "maximum": 100,
                "type": "integer",
                "title": "My field",
                "description": "My great field",
                "factory": "Integer",
                "default": 50,
            },
            adapter.get_schema(),
        )

    def test_choice(self):
        field = schema.Choice(
            __name__="myfield",
            title="My field",
            description="My great field",
            vocabulary=self.dummy_vocabulary,
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "string",
                "title": "My field",
                "description": "My great field",
                "factory": "Choice",
                "enum": ["foo", "bar"],
                "enumNames": ["Foo", "Bar"],
                "choices": [("foo", "Foo"), ("bar", "Bar")],
                "vocabulary": {"@id": "http://nohost/plone/@sources/myfield"},
            },
            adapter.get_schema(),
        )

    def test_choice_inline_array(self):
        field = schema.Choice(
            __name__="myfield",
            title="My field",
            description="My great field",
            values=["foo", "bar"],
        )

        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "string",
                "title": "My field",
                "description": "My great field",
                "factory": "Choice",
                "enum": ["foo", "bar"],
                "enumNames": [None, None],
                "choices": [("foo", None), ("bar", None)],
                "vocabulary": {"@id": "http://nohost/plone/@sources/myfield"},
            },
            adapter.get_schema(),
        )

    def test_choice_named_vocab(self):
        field = schema.Choice(
            title="My field",
            description="My great field",
            vocabulary="plone.app.vocabularies.ReallyUserFriendlyTypes",
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "string",
                "title": "My field",
                "description": "My great field",
                "factory": "Choice",
                "vocabulary": {
                    "@id": "http://nohost/plone/@vocabularies/plone.app.vocabularies.ReallyUserFriendlyTypes"
                },  # noqa
            },
            adapter.get_schema(),
        )

    def test_choice_source_vocab(self):
        field = schema.Choice(
            __name__="myfield",
            title="My field",
            description="My great field",
            source=self.dummy_source_vocab,
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "string",
                "title": "My field",
                "description": "My great field",
                "factory": "Choice",
                "enum": ["foo", "bar"],
                "enumNames": ["Foo", "Bar"],
                "choices": [("foo", "Foo"), ("bar", "Bar")],
                "vocabulary": {"@id": "http://nohost/plone/@sources/myfield"},
            },
            adapter.get_schema(),
        )

    def test_collection(self):
        field = schema.List(
            title="My field",
            description="My great field",
            min_length=1,
            value_type=schema.TextLine(
                title="Text", description="Text field", default="Default text"
            ),
            default=["foobar"],
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "array",
                "title": "My field",
                "description": "My great field",
                "factory": "List",
                "default": ["foobar"],
                "minItems": 1,
                "uniqueItems": False,
                "additionalItems": True,
                "items": {
                    "type": "string",
                    "title": "Text",
                    "description": "Text field",
                    "factory": "Text line (String)",
                    "default": "Default text",
                },
            },
            adapter.get_schema(),
        )

        # Test Tuple
        field = schema.Tuple(title="My field", value_type=schema.Int(), default=(1, 2))
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "array",
                "title": "My field",
                "description": "",
                "factory": "Tuple",
                "uniqueItems": True,
                "additionalItems": True,
                "items": {
                    "title": "",
                    "description": "",
                    "type": "integer",
                    "factory": "Integer",
                },
                "default": (1, 2),
            },
            adapter.get_schema(),
        )

        # Test Set
        field = schema.Set(title="My field", value_type=schema.TextLine())
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "array",
                "title": "My field",
                "description": "",
                "factory": "Multiple Choice",
                "uniqueItems": True,
                "additionalItems": True,
                "items": {
                    "title": "",
                    "description": "",
                    "factory": "Text line (String)",
                    "type": "string",
                },
            },
            adapter.get_schema(),
        )

        # List of choices
        field = schema.List(
            __name__="myfield",
            title="My field",
            value_type=schema.Choice(vocabulary=self.dummy_vocabulary),
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "array",
                "title": "My field",
                "description": "",
                "factory": "List",
                "uniqueItems": True,
                "additionalItems": True,
                "items": {
                    "title": "",
                    "description": "",
                    "factory": "Choice",
                    "type": "string",
                    "enum": ["foo", "bar"],
                    "enumNames": ["Foo", "Bar"],
                    "choices": [("foo", "Foo"), ("bar", "Bar")],
                    "vocabulary": {"@id": "http://nohost/plone/@sources/"},
                },
            },
            adapter.get_schema(),
        )

    def test_object(self):
        field = schema.Object(
            title="My field", description="My great field", schema=IDummySchema
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "object",
                "title": "My field",
                "description": "My great field",
                "factory": "File",
                "properties": {
                    "field1": {
                        "title": "Foo",
                        "description": "",
                        "factory": "Yes/No",
                        "type": "boolean",
                    },
                    "field2": {
                        "title": "Bar",
                        "description": "",
                        "factory": "Text line (String)",
                        "type": "string",
                    },
                },
            },
            adapter.get_schema(),
        )

    def test_richtext(self):
        field = RichText(title="My field", description="My great field")
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "string",
                "title": "My field",
                "factory": "Rich Text",
                "description": "My great field",
                "widget": "richtext",
            },
            adapter.get_schema(),
        )

    def test_date(self):
        field = schema.Date(
            title="My field", description="My great field", default=date(2016, 1, 1)
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "string",
                "title": "My field",
                "factory": "Date",
                "description": "My great field",
                "default": date(2016, 1, 1),
                "widget": "date",
            },
            adapter.get_schema(),
        )

    def test_datetime(self):
        field = schema.Datetime(title="My field", description="My great field")
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "string",
                "title": "My field",
                "factory": "Date/Time",
                "description": "My great field",
                "widget": "datetime",
            },
            adapter.get_schema(),
        )

    def test_jsonfield(self):
        field = JSONField(
            title="My field", description="My great field", widget="my_widget_name"
        )
        adapter = getMultiAdapter(
            (field, self.portal, self.request), IJsonSchemaProvider
        )

        self.assertEqual(
            {
                "type": "dict",
                "title": "My field",
                "factory": "JSONField",
                "description": "My great field",
                "widget": "my_widget_name",
            },
            adapter.get_schema(),
        )



from itertools import chain
from plone.app.workflow.browser.sharing import merge_search_results
from plone.restapi.interfaces import ISerializeToJsonSummary
from plone.restapi.services import Service
from Products.CMFCore.utils import getToolByName
from zExceptions import BadRequest
from zope.component import getMultiAdapter


class PrincipalsGet(Service):
    def reply(self):
        if self.request.form.get("search", False):
            self.search_term = self.request.form["search"]
        else:
            raise BadRequest('Required "search" parameter is missing.')

        users = self.serialize_principals(self.user_search_results())
        groups = self.serialize_principals(self.group_search_results())

        principals = dict(users=users, groups=groups)

        return principals

    def serialize_principals(self, principals):
        result = []
        for principal in principals:
            serializer = getMultiAdapter(
                (principal, self.request), ISerializeToJsonSummary
            )
            result.append(serializer())
        return result

    def user_search_results(self):
        def search_for_principal(hunter, search_term):
            return merge_search_results(
                chain(
                    *(
                        hunter.searchUsers(**{field: search_term})
                        for field in ["name", "fullname", "email"]
                    )
                ),
                "userid",
            )

        def get_principal_by_id(user_id):
            mtool = getToolByName(self.context, "portal_membership")
            return mtool.getMemberById(user_id)

        return self._principal_search_results(
            search_for_principal, get_principal_by_id, "user", "userid"
        )

    def group_search_results(self):
        def search_for_principal(hunter, search_term):
            return merge_search_results(
                chain(
                    *(
                        hunter.searchGroups(**{field: search_term})
                        for field in ["id", "title"]
                    )
                ),
                "groupid",
            )

        def get_principal_by_id(group_id):
            portal_groups = getToolByName(self.context, "portal_groups")
            return portal_groups.getGroupById(group_id)

        return self._principal_search_results(
            search_for_principal, get_principal_by_id, "group", "groupid"
        )

    def _principal_search_results(
        self, search_for_principal, get_principal_by_id, principal_type, id_key
    ):

        hunter = getMultiAdapter((self.context, self.request), name="pas_search")

        principals = []
        for principal_info in search_for_principal(hunter, self.search_term):
            principal_id = principal_info[id_key]
            principals.append(get_principal_by_id(principal_id))

        return principals



from plone.restapi.services import Service
from plone.restapi.services.locking import lock_info


class Lock(Service):
    """Lock information about the current lock"""

    def reply(self):
        return lock_info(self.context)



from plone.restapi.batching import HypermediaBatch
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.services.sources.get import get_field_by_name
from plone.restapi.services.sources.get import SourcesGet
from z3c.formwidget.query.interfaces import IQuerySource
from zope.component import getMultiAdapter
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse


@implementer(IPublishTraverse)
class QuerySourcesGet(SourcesGet):
    def reply(self):
        if len(self.params) != 1:
            return self._error(
                400, "Bad Request", "Must supply exactly one path parameter (fieldname)"
            )
        fieldname = self.params[0]

        field = get_field_by_name(fieldname, self.context)
        if field is None:
            return self._error(404, "Not Found", "No such field: %r" % fieldname)
        bound_field = field.bind(self.context)

        source = bound_field.source
        if not IQuerySource.providedBy(source):
            return self._error(
                404, "Not Found", "Field %r does not have an IQuerySource" % fieldname
            )

        if "query" not in self.request.form:
            return self._error(
                400,
                "Bad Request",
                "Enumerating querysources is not supported. Please search "
                "the source using the ?query= QS parameter",
            )

        query = self.request.form["query"]

        result = source.search(query)

        terms = []
        for term in result:
            terms.append(term)

        batch = HypermediaBatch(self.request, terms)

        serialized_terms = []
        for term in batch:
            serializer = getMultiAdapter(
                (term, self.request), interface=ISerializeToJson
            )
            serialized_terms.append(serializer())

        result = {
            "@id": batch.canonical_url,
            "items": serialized_terms,
            "items_total": batch.items_total,
        }
        links = batch.links
        if links:
            result["batching"] = links
        return result



from AccessControl.security import checkPermission
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import ISerializeToJsonSummary
from plone.restapi.services import Service
from plone.tiles.interfaces import ITileType
from zope.deprecation import deprecated
from zope.component import getMultiAdapter
from zope.component import getUtilitiesFor
from zope.component import getUtility
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse

import zope.deprecation
import warnings
import sys


sys.modules["plone.restapi.services.tiles"] = deprecated(
    zope.deprecation,
    "``plone.restapi.services.tiles`` is deprecated and will be removed in plone.restapi 9.",
)


@implementer(IPublishTraverse)
class TilesGet(Service):
    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        # Treat any path segments after /@types as parameters
        self.params.append(name)
        return self

    def reply(self):
        warnings.warn(
            "``plone.restapi.services.tiles`` is deprecated and will be removed in plone.restapi 9.",
            DeprecationWarning,
        )

        if self.params and len(self.params) > 0:
            self.content_type = "application/json+schema"
            try:
                tile = getUtility(ITileType, name=self.params[0])
                return getMultiAdapter((tile, self.request), ISerializeToJson)()
            except KeyError:
                self.content_type = "application/json"
                self.request.response.setStatus(404)
                return {
                    "type": "NotFound",
                    "message": f'Tile "{self.params[0]}" could not be found.',
                }

        result = []
        tiles = getUtilitiesFor(ITileType, context=self.context)
        for name, tile in tiles:
            serializer = getMultiAdapter((tile, self.request), ISerializeToJsonSummary)
            if checkPermission(tile.add_permission, self.context):
                result.append(serializer())

        return result



from AccessControl.SecurityManagement import getSecurityManager
from AccessControl.SecurityManagement import newSecurityManager
from AccessControl.SecurityManagement import setSecurityManager
from AccessControl.users import UnrestrictedUser as BaseUnrestrictedUser
from contextlib import contextmanager
from zope.component import getMultiAdapter


class UnrestrictedUser(BaseUnrestrictedUser):
    """Unrestricted user that still has an id."""

    def getId(self):
        """Return the ID of the user."""
        return self.getUserName()


@contextmanager
def elevated_privileges(context):
    """Temporarily elevate current user's privileges.

    See http://docs.plone.org/develop/plone/security/permissions.html
    for more documentation on this code.

    """
    sm = getSecurityManager()
    try:
        portal = getMultiAdapter(
            (context, context.REQUEST), name="plone_portal_state"
        ).portal()
        tmp_user = UnrestrictedUser(sm.getUser().getId(), "", ("manage", "Manager"), "")
        tmp_user = tmp_user.__of__(portal.acl_users)
        newSecurityManager(None, tmp_user)

        yield
    finally:
        setSecurityManager(sm)



from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.dexterity.interfaces import IDexterityFTI
from plone.dexterity.utils import createContentInContainer
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from zope.component import queryUtility

import transaction
import unittest


class TestContentBlocks(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        fti = queryUtility(IDexterityFTI, name="Document")
        behavior_list = [a for a in fti.behaviors]
        behavior_list.append("volto.blocks")
        behavior_list.append("plone.leadimage")
        fti.behaviors = tuple(behavior_list)

        self.doc = createContentInContainer(
            self.portal, "Document", id="doc", title="A document"
        )
        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_patch_blocks_list(self):
        response = self.api_session.patch(
            "/doc",
            json={
                "blocks": {
                    "uuid1": {"@type": "title"},
                    "uuid2": {"@type": "description"},
                }
            },
        )

        self.assertEqual(response.status_code, 204)

        response = self.api_session.get("/doc")
        response = response.json()

        self.assertEqual(
            response["blocks"],
            {"uuid1": {"@type": "title"}, "uuid2": {"@type": "description"}},
        )

    def test_patch_blocks_layout(self):
        response = self.api_session.patch(
            "/doc", json={"blocks_layout": {"items": ["#uuid1", "#uuid2"]}}
        )

        self.assertEqual(response.status_code, 204)

        response = self.api_session.get("/doc")
        response = response.json()

        self.assertEqual(response["blocks_layout"], {"items": ["#uuid1", "#uuid2"]})

    def test_get_blocks_layout_schema(self):
        response = self.api_session.get("/@types/Document")

        self.assertEqual(response.status_code, 200)
        response = response.json()

    # These are not failing because the patch operations doesn't validate
    # fields right now
    # def test_patch_blocks_list_wrong_type(self):
    #     response = self.api_session.patch(
    #         '/doc',
    #         json={
    #             "blocks": [{'uuid1': {'@type': 'title'}}]
    #         })
    #     self.assertEqual(response.status_code, 500)

    # def test_patch_blocks_layout_wrong_type(self):
    #     response = self.api_session.patch(
    #         '/doc',
    #         json={
    #             "blocks_layout": {'uuid1': {'@type': 'title'}}
    #         })
    #     self.assertEqual(response.status_code, 500)



from Acquisition import aq_parent
from plone.restapi.deserializer import json_body
from plone.restapi.services import Service
from Products.CMFCore.utils import getToolByName
from zExceptions import BadRequest
from zope.component import getMultiAdapter
from zope.interface import alsoProvides
from zope.security import checkPermission

import plone


class BaseCopyMove(Service):
    """Base service for copy/move operations."""

    def __init__(self, context, request):
        super().__init__(context, request)
        self.portal = getMultiAdapter(
            (self.context, self.request), name="plone_portal_state"
        ).portal()
        self.portal_url = self.portal.absolute_url()
        self.catalog = getToolByName(self.context, "portal_catalog")

    def get_object(self, key):
        """Get an object by url, path or UID."""
        if key.startswith(self.portal_url):
            # Resolve by URL
            key = key[len(self.portal_url) + 1 :]
            return self.portal.restrictedTraverse(key, None)
        elif key.startswith("/"):
            # Resolve by path
            return self.portal.restrictedTraverse(key.lstrip("/"), None)
        else:
            # Resolve by UID
            brain = self.catalog(UID=key)
            if brain:
                return brain[0].getObject()

    def reply(self):
        # return 401/403 Forbidden if the user has no permission
        if not checkPermission("cmf.AddPortalContent", self.context):
            pm = getToolByName(self.context, "portal_membership")
            if bool(pm.isAnonymousUser()):
                self.request.response.setStatus(401)
            else:
                self.request.response.setStatus(403)
            return

        data = json_body(self.request)

        source = data.get("source", None)

        if not source:
            raise BadRequest("Property 'source' is required")

        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        if not isinstance(source, list):
            source = [source]

        parents_ids = {}
        for item in source:
            obj = self.get_object(item)
            if obj is not None:
                if self.is_moving:
                    # To be able to safely move the object, the user requires
                    # permissions on the parent
                    if not checkPermission(
                        "zope2.DeleteObjects", obj
                    ) and not checkPermission("zope2.DeleteObjects", aq_parent(obj)):
                        self.request.response.setStatus(403)
                        return
                parent = aq_parent(obj)
                if parent in parents_ids:
                    parents_ids[parent].append(obj.getId())
                else:
                    parents_ids[parent] = [obj.getId()]

        results = []
        for parent, ids in parents_ids.items():
            result = self.context.manage_pasteObjects(
                cb_copy_data=self.clipboard(parent, ids)
            )
            for res in result:
                results.append(
                    {
                        "source": "{}/{}".format(parent.absolute_url(), res["id"]),
                        "target": "{}/{}".format(
                            self.context.absolute_url(), res["new_id"]
                        ),
                    }
                )
        return results

    def clipboard(self, parent, ids):
        """Get clipboard data"""
        raise NotImplementedError


class Copy(BaseCopyMove):
    """Copies existing content objects."""

    is_moving = False

    def clipboard(self, parent, ids):
        return parent.manage_copyObjects(ids=ids)


class Move(BaseCopyMove):
    """Moves existing content objects."""

    is_moving = True

    def clipboard(self, parent, ids):
        return parent.manage_cutObjects(ids=ids)



from plone import api
from zope.component import queryUtility
from plone.dexterity.interfaces import IDexterityFTI

import logging

logger = logging.getLogger(__name__)

DEPRECATED_NEW_BEHAVIOR_NAME = "plone.restapi.behaviors.IBlocks"
SHORT_NEW_NAME = "volto.blocks"


def rename_iface_to_name_in_blocks_behavior(setup_context):
    """Rename iface name to the short name in blocks"""
    pt = api.portal.get_tool("portal_types")

    for _type in pt.objectIds():
        fti = queryUtility(IDexterityFTI, name=_type)
        if fti and DEPRECATED_NEW_BEHAVIOR_NAME in fti.behaviors:
            new_fti = [
                currentbehavior
                for currentbehavior in fti.behaviors
                if currentbehavior != DEPRECATED_NEW_BEHAVIOR_NAME
            ]
            new_fti.append(SHORT_NEW_NAME)
            fti.behaviors = tuple(new_fti)
            logger.info(f"Migrated behavior of {_type} type")



from AccessControl import getSecurityManager
from plone.app.contenttypes.interfaces import ILink
from plone.app.contenttypes.utils import replace_link_variables_by_paths
from plone.app.textfield.interfaces import IRichText
from plone.dexterity.interfaces import IDexterityContent
from plone.namedfile.interfaces import INamedFileField
from plone.namedfile.interfaces import INamedImageField
from plone.restapi.imaging import get_original_image_url
from plone.restapi.imaging import get_scales
from plone.restapi.interfaces import IFieldSerializer
from plone.restapi.interfaces import IPrimaryFieldTarget
from plone.restapi.serializer.converters import json_compatible
from plone.restapi.serializer.utils import uid_to_url
from Products.CMFCore.permissions import ModifyPortalContent
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.interface import implementer
from zope.interface import Interface
from zope.schema.interfaces import IChoice
from zope.schema.interfaces import ICollection
from zope.schema.interfaces import IField
from zope.schema.interfaces import ITextLine
from zope.schema.interfaces import IVocabularyTokenized

import logging


log = logging.getLogger(__name__)


@adapter(IField, IDexterityContent, Interface)
@implementer(IFieldSerializer)
class DefaultFieldSerializer:
    def __init__(self, field, context, request):
        self.context = context
        self.request = request
        self.field = field

    def __call__(self):
        return json_compatible(self.get_value())

    def get_value(self, default=None):
        return getattr(self.field.interface(self.context), self.field.__name__, default)


@adapter(IChoice, IDexterityContent, Interface)
@implementer(IFieldSerializer)
class ChoiceFieldSerializer(DefaultFieldSerializer):
    def __call__(self):
        # Binding is necessary for named vocabularies
        if IField.providedBy(self.field):
            self.field = self.field.bind(self.context)
        value = self.get_value()
        if value is not None and IVocabularyTokenized.providedBy(self.field.vocabulary):
            try:
                term = self.field.vocabulary.getTerm(value)
                value = {"token": term.token, "title": term.title}
            # Some fields (e.g. language) have a default value that is not in
            # vocabulary
            except LookupError:
                pass
        return json_compatible(value)


@adapter(ICollection, IDexterityContent, Interface)
@implementer(IFieldSerializer)
class CollectionFieldSerializer(DefaultFieldSerializer):
    def __call__(self):
        # Binding is necessary for named vocabularies
        if IField.providedBy(self.field):
            self.field = self.field.bind(self.context)
        value = self.get_value()
        value_type = self.field.value_type
        if (
            value is not None
            and IChoice.providedBy(value_type)
            and IVocabularyTokenized.providedBy(value_type.vocabulary)
        ):
            values = []
            for v in value:
                try:
                    term = value_type.vocabulary.getTerm(v)
                    values.append({"token": term.token, "title": term.title})
                except LookupError:
                    log.warning("Term lookup error: %r" % v)
            value = values
        return json_compatible(value)


@adapter(INamedImageField, IDexterityContent, Interface)
class ImageFieldSerializer(DefaultFieldSerializer):
    def __call__(self):
        image = self.field.get(self.context)
        if not image:
            return

        width, height = image.getImageSize()

        url = get_original_image_url(self.context, self.field.__name__, width, height)

        scales = get_scales(self.context, self.field, width, height)
        result = {
            "filename": image.filename,
            "content-type": image.contentType,
            "size": image.getSize(),
            "download": url,
            "width": width,
            "height": height,
            "scales": scales,
        }
        return json_compatible(result)


@adapter(INamedFileField, IDexterityContent, Interface)
class FileFieldSerializer(DefaultFieldSerializer):
    def __call__(self):
        namedfile = self.field.get(self.context)
        if namedfile is None:
            return

        url = "/".join((self.context.absolute_url(), "@@download", self.field.__name__))
        result = {
            "filename": namedfile.filename,
            "content-type": namedfile.contentType,
            "size": namedfile.getSize(),
            "download": url,
        }
        return json_compatible(result)


@adapter(IRichText, IDexterityContent, Interface)
class RichttextFieldSerializer(DefaultFieldSerializer):
    def __call__(self):
        value = self.get_value()
        return json_compatible(value, self.context)


@adapter(ITextLine, ILink, Interface)
class TextLineFieldSerializer(DefaultFieldSerializer):
    def __call__(self):
        if self.field.getName() != "remoteUrl":
            return super().__call__()
        value = self.get_value()

        # Expect that all internal links will have resolveuid
        if value and "resolveuid" in value:
            return uid_to_url(value)

        # Fallback in case we still have a variable in there
        path = replace_link_variables_by_paths(context=self.context, url=value)
        portal = getMultiAdapter(
            (self.context, self.context.REQUEST), name="plone_portal_state"
        ).portal()
        # We should traverse unrestricted, just in case that the path to the object
        # is not all public, we should be able to reach it by finger pointing
        ref_obj = portal.unrestrictedTraverse(path, None)
        if ref_obj:
            value = ref_obj.absolute_url()
            return json_compatible(value)
        else:
            # The URL does not point to an existing object, so just return the value
            # without value interpolation
            return json_compatible(value.replace("${portal_url}", ""))


@adapter(IField, IDexterityContent, Interface)
@implementer(IPrimaryFieldTarget)
class DefaultPrimaryFieldTarget:
    def __init__(self, field, context, request):
        self.context = context
        self.request = request
        self.field = field

    def use_primary_field_target(self):
        sm = getSecurityManager()
        perm = bool(sm.checkPermission(ModifyPortalContent, self.context))
        if perm:
            return False
        return True

    def __call__(self):
        return


@adapter(INamedFileField, IDexterityContent, Interface)
class PrimaryFileFieldTarget(DefaultPrimaryFieldTarget):
    def __call__(self):
        if not self.use_primary_field_target():
            return

        namedfile = self.field.get(self.context)
        if namedfile is None:
            return

        return "/".join(
            (self.context.absolute_url(), "@@download", self.field.__name__)
        )



from plone.restapi.services import Service
from zope.interface import alsoProvides
from plone.app.iterate.interfaces import ICheckinCheckoutPolicy
from plone.app.iterate.interfaces import IWCContainerLocator
from Acquisition import aq_inner
from zope.component import getAdapters
from zope.component import getMultiAdapter
from Products.CMFCore.utils import getToolByName

import plone.protect.interfaces

# It seems that p.a.iterate allows to locate the WC in the user folder,
# for now, ignore it and use always the location as the parent
# but allow space to implement it in the future if it's still relevant
WC_LOCATION_MODE = "plone.app.iterate.parent"


class CreateWorkingCopy(Service):
    def reply(self):
        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        control = getMultiAdapter((self.context, self.request), name="iterate_control")
        if not control.checkout_allowed():
            pm = getToolByName(self.context, "portal_membership")
            if bool(pm.isAnonymousUser()):
                return self._error(401, "Not authenticated", "Checkout not allowed")
            else:
                return self._error(403, "Not authorized", "Checkout not allowed")

        locator = None
        try:
            locator = [
                c["locator"] for c in self.containers() if c["name"] == WC_LOCATION_MODE
            ][0]
        except IndexError:
            return self._error(
                500, "InternalServerError", "Cannot find checkout location"
            )

        policy = ICheckinCheckoutPolicy(self.context)
        wc = policy.checkout(locator())

        self.request.response.setStatus(201)
        self.request.response.setHeader("Location", self.context.absolute_url())
        return {"@id": wc.absolute_url()}

    def containers(self):
        """Get a list of potential containers (copied over from p.a.iterate)"""
        context = aq_inner(self.context)
        for name, locator in getAdapters((context,), IWCContainerLocator):
            if locator.available:
                yield dict(name=name, locator=locator)

    def _error(self, status, type, message):
        self.request.response.setStatus(status)
        return {"error": {"type": type, "message": message}}



from zExceptions import BadRequest
from zope.component import adapter
from zope.component import queryMultiAdapter
from zope.interface import alsoProvides
from zope.interface import implementer
from zope.interface import Interface
from plone.i18n.normalizer import idnormalizer
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import IDeserializeFromJson
from plone.restapi.deserializer import json_body
from plone.restapi.controlpanels import RegistryConfigletPanel
from plone.restapi.controlpanels.interfaces import IDexterityTypesControlpanel
import plone.protect.interfaces
from zope.interface import noLongerProvides
from plone.restapi.interfaces import IPloneRestapiLayer


@adapter(Interface, IPloneRestapiLayer)
@implementer(IDexterityTypesControlpanel)
class DexterityTypesControlpanel(RegistryConfigletPanel):
    schema = Interface
    configlet_id = "dexterity-types"
    configlet_category_id = "plone-content"

    def add(self, names):
        data = json_body(self.request)

        title = data.get("title", None)
        if not title:
            raise BadRequest("Property 'title' is required")

        tid = data.get("id", None)
        if not tid:
            tid = idnormalizer.normalize(title).replace("-", "_")

        description = data.get("description", "")

        properties = {"id": tid, "title": title, "description": description}

        # Disable CSRF protection
        if "IDisableCSRFProtection" in dir(plone.protect.interfaces):
            alsoProvides(self.request, plone.protect.interfaces.IDisableCSRFProtection)

        if IPloneRestapiLayer.providedBy(self.request):
            noLongerProvides(self.request, IPloneRestapiLayer)

        context = queryMultiAdapter(
            (self.context, self.request), name="dexterity-types"
        )
        add_type = queryMultiAdapter((context, self.request), name="add-type")
        fti = add_type.form_instance.create(data=properties)
        add_type.form_instance.add(fti)
        return self.get([tid])

    def get(self, names):
        name = names[0]

        if IPloneRestapiLayer.providedBy(self.request):
            noLongerProvides(self.request, IPloneRestapiLayer)

        context = queryMultiAdapter(
            (self.context, self.request), name="dexterity-types"
        )
        context = context.publishTraverse(self.request, name)
        serializer = ISerializeToJson(self)
        return serializer(context)

    def update(self, names):
        name = names[0]

        if IPloneRestapiLayer.providedBy(self.request):
            noLongerProvides(self.request, IPloneRestapiLayer)

        context = queryMultiAdapter(
            (self.context, self.request), name="dexterity-types"
        )
        context = context.publishTraverse(self.request, name)
        deserializer = IDeserializeFromJson(self)
        return deserializer(context)

    def delete(self, names):
        name = names[0]

        if IPloneRestapiLayer.providedBy(self.request):
            noLongerProvides(self.request, IPloneRestapiLayer)

        context = queryMultiAdapter(
            (self.context, self.request), name="dexterity-types"
        )
        edit = queryMultiAdapter((context, self.request), name="edit")
        edit.form_instance.remove((name, None))



from Acquisition import aq_parent
from plone.restapi.services.content.utils import add
from plone.restapi.services.content.utils import create
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from Products.CMFPlone.interfaces import ISelectableConstrainTypes
from zExceptions import Unauthorized
from zope.component import getGlobalSiteManager
from zope.event import notify
from zope.lifecycleevent import ObjectCreatedEvent
from zope.lifecycleevent.interfaces import IObjectAddedEvent

import unittest


class TestCreateContent(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

        self.folder = self.portal[
            self.portal.invokeFactory("Folder", id="folder", title="My Folder")
        ]

    def test_create_content_with_provided_id(self):
        obj = create(self.folder, "Document", "my-document")
        self.assertEqual(obj.portal_type, "Document")
        self.assertEqual(obj.getId(), "my-document")

    def test_create_content_without_provided_id(self):
        obj = create(self.folder, "Document")
        self.assertEqual(obj.portal_type, "Document")
        self.assertTrue(obj.getId().startswith("document."))

    def test_create_content_without_add_permission_raises_unauthorized(self):
        self.folder.manage_permission(
            "plone.app.contenttypes: Add Document", [], acquire=False
        )
        with self.assertRaises(Unauthorized):
            create(self.folder, "Document", "my-document")

    def test_create_of_disallowed_content_type_raises_unauthorized(self):
        self.portal.portal_types.Folder.filter_content_types = True
        self.portal.portal_types.Folder.allowed_content_types = ()
        with self.assertRaises(Unauthorized):
            create(self.folder, "Document", "my-document")

    def test_create_of_constrained_content_type_raises_unauthorized(self):
        constrains = ISelectableConstrainTypes(self.folder)
        constrains.setConstrainTypesMode(1)
        constrains.setLocallyAllowedTypes(["File"])
        with self.assertRaises(Unauthorized):
            create(self.folder, "Document", "my-document")


class TestAddContent(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

        self.folder = self.portal[
            self.portal.invokeFactory("Folder", id="folder", title="My Folder")
        ]

    def test_add_content_to_container(self):
        obj = create(self.folder, "Document", "my-document")
        obj = add(self.folder, obj)
        self.assertEqual(aq_parent(obj), self.folder)

    def test_add_content_to_container_renames_id(self):
        obj = create(self.folder, "Document", title="My Document")
        obj = add(self.folder, obj)
        self.assertEqual(obj.getId(), "my-document")

    def test_add_content_to_container_keeps_id(self):
        obj = create(self.folder, "Document", "doc-1", "My Document")
        obj = add(self.folder, obj, rename=False)
        self.assertEqual(obj.getId(), "doc-1")

    def test_add_content_to_container_and_move_on_added_event(self):
        sm = getGlobalSiteManager()

        def move_object(event):
            self.portal.manage_pasteObjects(
                cb_copy_data=self.folder.manage_cutObjects(ids=["my-document"])
            )

        sm.registerHandler(move_object, (IObjectAddedEvent,))

        obj = create(self.folder, "Document", "my-document")
        notify(ObjectCreatedEvent(obj))
        obj = add(self.folder, obj)
        self.assertEqual(aq_parent(obj), self.portal)

        sm.unregisterHandler(move_object, (IObjectAddedEvent,))



from AccessControl import allow_module
from AccessControl.Permissions import add_user_folders
from plone.restapi.pas import plugin
from Products.PluggableAuthService.PluggableAuthService import registerMultiPlugin
from zope.i18nmessageid import MessageFactory


_ = MessageFactory("plone.restapi")
PROJECT_NAME = "plone.restapi"


allow_module("json")


def initialize(context):
    registerMultiPlugin(plugin.JWTAuthenticationPlugin.meta_type)
    context.registerClass(
        plugin.JWTAuthenticationPlugin,
        permission=add_user_folders,
        constructors=(
            plugin.manage_addJWTAuthenticationPlugin,
            plugin.addJWTAuthenticationPlugin,
        ),
        visibility=None,
    )



from plone.folder.interfaces import IExplicitOrdering
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.interfaces.siteroot import IPloneSiteRoot
from zExceptions import BadRequest


class OrderingMixin:
    def handle_ordering(self, data):
        if "ordering" in data:
            obj_id = data["ordering"]["obj_id"]
            delta = data["ordering"]["delta"]
            subset_ids = data["ordering"].get("subset_ids")

            # The REST api returns only content items and a Zope resource
            # may contain non-content items. We need to set the subset_ids
            # so we'll move items relative to each other.
            if not subset_ids:
                subset_ids = self.context.contentIds()

            self.reorderItems(obj_id, delta, subset_ids)

        if "sort" in data:
            # resort folder items with given parameters
            sort_on = data["sort"]["on"]
            sort_order = data["sort"]["order"]
            self.resortAllItemsInContext(sort_on, sort_order)

    def reorderItems(self, obj_id, delta, subset_ids):
        # Based on wildcard.foldercontents.viewsItemOrder
        ordering = self.getOrdering()
        if ordering is None:
            msg = "Content ordering is not supported by this resource"
            raise BadRequest(msg)

        # Make sure we're seeing the same order as the client is.
        if subset_ids:
            position_id = [(ordering.getObjectPosition(i), i) for i in subset_ids]
            position_id.sort()
            if subset_ids != [i for position, i in position_id]:
                raise BadRequest("Client/server ordering mismatch")

        # All movement is relative to the subset of ids, if passed in.
        if delta == "top":
            ordering.moveObjectsToTop([obj_id], subset_ids=subset_ids)
        elif delta == "bottom":
            ordering.moveObjectsToBottom([obj_id], subset_ids=subset_ids)
        else:
            delta = int(delta)
            ordering.moveObjectsByDelta([obj_id], delta, subset_ids=subset_ids)

    def resortAllItemsInContext(self, sort_on, sort_order):
        # Based on plone.app.content.browser.contents.rearrange
        ordering = self.getOrdering()
        if ordering is None:
            msg = "Content ordering is not supported by this resource"
            raise BadRequest(msg)

        catalog = getToolByName(self.context, "portal_catalog")
        query = {
            "path": {"query": "/".join(self.context.getPhysicalPath()), "depth": 1},
            "sort_on": sort_on,
            "show_inactive": True,
        }
        brains = catalog(**query)
        if sort_order in ("reverse", "descending"):
            brains = [b for b in reversed(brains)]
        for idx, brain in enumerate(brains):
            ordering.moveObjectToPosition(brain.id, idx)

    def getOrdering(self):
        if IPloneSiteRoot.providedBy(self.context):
            return self.context
        if getattr(self.context, "getOrdering", None):
            ordering = self.context.getOrdering()
            if IExplicitOrdering.providedBy(ordering):
                return ordering



def assign_use_api_permission(setup_context):
    """Assign the 'plone.restapi: Use REST API' permission to Anonymous."""
    setup_context.runImportStepFromProfile(
        "profile-plone.restapi.upgrades:0002",
        "rolemap",
        run_dependencies=False,
        purge_old=False,
    )



from .mixins import OrderingMixin
from AccessControl import getSecurityManager
from plone.autoform.interfaces import WRITE_PERMISSIONS_KEY
from plone.dexterity.interfaces import IDexterityContent
from plone.dexterity.utils import iterSchemata
from plone.restapi.deserializer import json_body
from plone.restapi.interfaces import IDeserializeFromJson
from plone.restapi.interfaces import IFieldDeserializer
from plone.supermodel.utils import mergedTaggedValueDict
from z3c.form.interfaces import IDataManager
from z3c.form.interfaces import IManagerValidator
from zExceptions import BadRequest
from zope.component import adapter
from zope.component import queryMultiAdapter
from zope.component import queryUtility
from zope.event import notify
from zope.interface import implementer
from zope.interface import Interface
from zope.lifecycleevent import Attributes
from zope.lifecycleevent import ObjectModifiedEvent
from zope.schema import getFields
from zope.schema.interfaces import ValidationError
from zope.security.interfaces import IPermission


@implementer(IDeserializeFromJson)
@adapter(IDexterityContent, Interface)
class DeserializeFromJson(OrderingMixin):
    def __init__(self, context, request):
        self.context = context
        self.request = request

        self.sm = getSecurityManager()
        self.permission_cache = {}
        self.modified = {}

    def __call__(
        self, validate_all=False, data=None, create=False, mask_validation_errors=True
    ):  # noqa: ignore=C901

        if data is None:
            data = json_body(self.request)

        schema_data, errors = self.get_schema_data(data, validate_all, create)

        # Validate schemata
        for schema, field_data in schema_data.items():
            validator = queryMultiAdapter(
                (self.context, self.request, None, schema, None), IManagerValidator
            )
            for error in validator.validate(field_data):
                errors.append({"error": error, "message": str(error)})

        if errors:
            if mask_validation_errors:
                # Drop Python specific error classes in order to be able to better handle
                # errors on front-end
                for error in errors:
                    error["error"] = "ValidationError"
            raise BadRequest(errors)

        # We'll set the layout after the validation and even if there
        # are no other changes.
        if "layout" in data:
            layout = data["layout"]
            self.context.setLayout(layout)

        # OrderingMixin
        self.handle_ordering(data)

        if self.modified and not create:
            descriptions = []
            for interface, names in self.modified.items():
                descriptions.append(Attributes(interface, *names))
            notify(ObjectModifiedEvent(self.context, *descriptions))

        return self.context

    def get_schema_data(self, data, validate_all, create=False):
        schema_data = {}
        errors = []

        for schema in iterSchemata(self.context):
            write_permissions = mergedTaggedValueDict(schema, WRITE_PERMISSIONS_KEY)

            for name, field in getFields(schema).items():
                __traceback_info__ = f"field={field}"

                field_data = schema_data.setdefault(schema, {})

                if field.readonly:
                    continue

                if name in data:
                    dm = queryMultiAdapter((self.context, field), IDataManager)
                    if not dm.canWrite():
                        continue

                    if not self.check_permission(write_permissions.get(name)):
                        continue

                    # set the field to missing_value if we receive null
                    if data[name] is None:
                        if not field.required:
                            if dm.get():
                                self.mark_field_as_changed(schema, name)
                            dm.set(field.missing_value)
                        else:
                            errors.append(
                                {
                                    "field": field.__name__,
                                    "message": (
                                        "{} is a required field.".format(
                                            field.__name__
                                        ),
                                        "Setting it to null is not allowed.",
                                    ),
                                }
                            )
                        continue

                    # Deserialize to field value
                    deserializer = queryMultiAdapter(
                        (field, self.context, self.request), IFieldDeserializer
                    )
                    if deserializer is None:
                        continue

                    try:
                        value = deserializer(data[name])
                    except ValueError as e:
                        errors.append({"message": str(e), "field": name, "error": e})
                    except ValidationError as e:
                        errors.append({"message": e.doc(), "field": name, "error": e})
                    else:
                        field_data[name] = value
                        current_value = dm.get()
                        if value != current_value:
                            should_change = True
                        elif create and dm.field.defaultFactory:
                            # During content creation we should set the value even if
                            # it is the same from the dm if the current_value was
                            # returned from a default_factory method
                            should_change = (
                                dm.field.defaultFactory(self.context) == current_value
                            )
                        else:
                            should_change = False

                        if should_change:
                            dm.set(value)
                            self.mark_field_as_changed(schema, name)

                elif validate_all:
                    # Never validate the changeNote of p.a.versioningbehavior
                    # The Versionable adapter always returns an empty string
                    # which is the wrong type. Should be unicode and should be
                    # fixed in p.a.versioningbehavior
                    if name == "changeNote":
                        continue
                    dm = queryMultiAdapter((self.context, field), IDataManager)
                    bound = field.bind(self.context)
                    try:
                        bound.validate(dm.get())
                    except ValidationError as e:
                        errors.append({"message": e.doc(), "field": name, "error": e})

        return schema_data, errors

    def mark_field_as_changed(self, schema, fieldname):
        """Collect the names of the modified fields. Use prefixed name because
        z3c.form does so.
        """

        prefixed_name = schema.__name__ + "." + fieldname
        self.modified.setdefault(schema, []).append(prefixed_name)

    def check_permission(self, permission_name):
        if permission_name is None:
            return True

        if permission_name not in self.permission_cache:
            permission = queryUtility(IPermission, name=permission_name)
            if permission is None:
                self.permission_cache[permission_name] = True
            else:
                self.permission_cache[permission_name] = bool(
                    self.sm.checkPermission(permission.title, self.context)
                )
        return self.permission_cache[permission_name]



from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession

import transaction
import unittest


class TestHistoryVersioning(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        self.portal.invokeFactory(
            "Document", id="doc_with_history", title="My Document"
        )
        self.doc = self.portal.doc_with_history
        self.doc.setTitle("Current version")

        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_response(self):
        response = self.api_session.get(self.doc.absolute_url())
        self.assertIn("version", response.json())



from AccessControl.interfaces import IRoleManager
from plone.restapi.deserializer import json_body
from plone.restapi.interfaces import IDeserializeFromJson
from Products.CMFCore.interfaces import ICatalogAware
from Products.CMFPlone.interfaces import IPloneSiteRoot
from zope.component import adapter
from zope.component import getMultiAdapter
from zope.event import notify
from zope.interface import implementer
from zope.interface import Interface


try:
    from plone.app.workflow.events import LocalrolesModifiedEvent

    LOCALROLES_MODIFIED_EVENT_AVAILABLE = True
except ImportError:
    # Plone < 4.3.4
    LOCALROLES_MODIFIED_EVENT_AVAILABLE = False


marker = object()


@implementer(IDeserializeFromJson)
@adapter(IRoleManager, Interface)
class DeserializeFromJson:
    """JSON deserializer for local roles"""

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def __call__(self):
        data = json_body(self.request)
        sharing_view = getMultiAdapter((self.context, self.request), name="sharing")

        # inherit roles
        inherit_reindex = False
        # block can be None, so we might get False or None, so we test
        # for a marker.
        inherit = data.get("inherit", marker)
        if inherit is not marker:
            inherit_reindex = sharing_view.update_inherit(status=inherit, reindex=False)
        # roles
        roles_reindex = False
        new_roles = data.get("entries", None)
        managed_roles = frozenset(r["id"] for r in sharing_view.roles())

        if new_roles is not None:
            # the roles are converted into a FrozenSet so we have to filter
            # the data structure we get.
            for user in new_roles:
                roles_list = [key for key in user["roles"] if user["roles"][key]]

                # Limit roles to ones the user is allowed to delegate
                roles_list = set(roles_list).intersection(managed_roles)

                user["roles"] = roles_list
            roles_reindex = sharing_view.update_role_settings(new_roles, reindex=False)

        # reindex object security
        can_reindex = ICatalogAware(self.context, None) or IPloneSiteRoot.providedBy(
            self.context
        )
        if can_reindex and (inherit_reindex or roles_reindex):
            self.context.reindexObjectSecurity()
            if LOCALROLES_MODIFIED_EVENT_AVAILABLE:
                notify(LocalrolesModifiedEvent(self.context, self.request))



from plone.restapi.controlpanels import IControlpanel
from plone.restapi.interfaces import IJsonCompatible
from plone.restapi.interfaces import ISerializeToJson
from plone.restapi.interfaces import ISerializeToJsonSummary
from plone.restapi.services import Service
from Products.CMFCore.utils import getToolByName
from zope.component import getAdapters
from zope.interface import implementer
from zope.publisher.interfaces import IPublishTraverse


@implementer(IPublishTraverse)
class ControlpanelsGet(Service):
    def __init__(self, context, request):
        super().__init__(context, request)
        self.params = []

    def publishTraverse(self, request, name):
        self.params.append(name)
        return self

    def get_controlpanel_adapters(self):
        adapters = getAdapters((self.context, self.request), provided=IControlpanel)
        for name, panel in adapters:
            panel.__name__ = name
            yield name, panel

    def available_controlpanels(self):
        panels = dict(self.get_controlpanel_adapters())
        panels_by_configlet = {p.configlet_id: name for name, p in panels.items()}

        pctool = getToolByName(self.context, "portal_controlpanel")
        for group in pctool.getGroups():
            for action_data in pctool.enumConfiglets(group=group["id"]):
                name = panels_by_configlet.get(action_data["id"])
                panel = panels.get(name)
                if panel:
                    yield panel

    def panel_by_name(self, name):
        panels = dict(self.get_controlpanel_adapters())
        return panels.get(name)

    def reply(self):
        if self.params:
            return self.reply_panel()

        def serialize(panels):
            for panel in panels:
                serializer = ISerializeToJsonSummary(panel)
                yield serializer()

        # List panels
        panels = self.available_controlpanels()
        return IJsonCompatible(list(serialize(panels)))

    def reply_panel(self):
        name = self.params[0]
        panel = self.panel_by_name(name)
        if panel is None:
            self.request.response.setStatus(404)
            return

        # Panel child request
        if len(self.params) > 1:
            return IJsonCompatible(panel.get(self.params[1:]))

        # Panel request
        return IJsonCompatible(ISerializeToJson(panel)())



from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.dexterity.interfaces import IDexterityFTI
from plone.dexterity.interfaces import IDexterityItem
from plone.dexterity.utils import createContentInContainer
from plone.restapi.behaviors import IBlocks
from plone.restapi.interfaces import IBlockSearchableText
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
from zope.component import adapter
from zope.component import provideAdapter
from zope.component import queryUtility
from zope.interface import implementer
from zope.publisher.interfaces.browser import IBrowserRequest

import transaction
import unittest


class TestSearchTextInBlocks(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

        fti = queryUtility(IDexterityFTI, name="Document")
        behavior_list = [a for a in fti.behaviors]
        behavior_list.append("volto.blocks")
        fti.behaviors = tuple(behavior_list)

        self.doc = createContentInContainer(
            self.portal, "Document", id="doc", title="A document"
        )
        transaction.commit()

    def tearDown(self):
        self.api_session.close()

    def test_search_text(self):
        response = self.api_session.patch(
            "/doc",
            json={
                "blocks_layout": {"items": ["uuid1", "uuid2"]},
                "blocks": {
                    "uuid1": {
                        "@type": "text",
                        "text": {
                            "blocks": [
                                {
                                    "data": {},
                                    "depth": 0,
                                    "entityRanges": [],
                                    "inlineStyleRanges": [],
                                    "key": "acv4f",
                                    "text": "Plone " "text " "for " "block ",
                                    "type": "unstyled",
                                }
                            ],
                            "entityMap": {},
                        },
                    },
                    "uuid2": {
                        "@type": "text",
                        "text": {
                            "blocks": [
                                {
                                    "data": {},
                                    "depth": 0,
                                    "entityRanges": [],
                                    "inlineStyleRanges": [],
                                    "key": "1m9qt",
                                    "text": "Volto " "text " "for " "block ",
                                    "type": "unstyled",
                                }
                            ],
                            "entityMap": {},
                        },
                    },
                },
            },
        )

        self.assertEqual(response.status_code, 204)

        query = {"SearchableText": "Volto", "metadata_fields": "Title"}
        response = self.api_session.get("/@search", params=query)
        json_response = response.json()
        self.assertEqual(json_response["items_total"], 1)
        self.assertEqual(json_response["items"][0]["Title"], "A document")

        query = {
            "SearchableText": "Plone",
            "metadata_fields": "Title",
            "portal_type": "Document",
        }
        response = self.api_session.get("/@search", params=query)
        json_response = response.json()
        self.assertEqual(json_response["items_total"], 1)
        self.assertEqual(json_response["items"][0]["Title"], "A document")

    def test_register_block_searchabletext(self):
        @implementer(IBlockSearchableText)
        @adapter(IBlocks, IBrowserRequest)
        class TestSearchableTextAdapter:
            def __init__(self, context, request):
                self.context = context
                self.request = request

            def __call__(self, value):

                return "discovered: %s" % value["text"]

        provideAdapter(
            TestSearchableTextAdapter,
            (IDexterityItem, IBrowserRequest),
            name="test_block",
        )

        blocks = {
            "uuid1": {
                "@type": "text",
                "text": {
                    "blocks": [
                        {
                            "data": {},
                            "depth": 0,
                            "entityRanges": [],
                            "inlineStyleRanges": [],
                            "key": "acv4f",
                            "text": "Plone " "text " "for " "block ",
                            "type": "unstyled",
                        }
                    ],
                    "entityMap": {},
                },
            },
            "uuid3": {"@type": "test_block", "text": "sample text"},
        }

        self.doc.blocks = blocks
        self.doc.blocks_layout = {"items": ["uuid1", "uuid3"]}
        from plone.indexer.interfaces import IIndexableObject
        from zope.component import queryMultiAdapter

        wrapper = queryMultiAdapter(
            (self.doc, self.portal.portal_catalog), IIndexableObject
        )

        assert "discovered: sample text" in wrapper.SearchableText

    def test_index_searchableText_value(self):
        response = self.api_session.patch(
            "/doc",
            json={
                "blocks_layout": {"items": ["uuid1", "uuid2"]},
                "blocks": {
                    "uuid1": {
                        "@type": "text",
                        "text": {
                            "blocks": [
                                {
                                    "data": {},
                                    "depth": 0,
                                    "entityRanges": [],
                                    "inlineStyleRanges": [],
                                    "key": "acv4f",
                                    "text": "Plone " "text " "for " "block ",
                                    "type": "unstyled",
                                }
                            ],
                            "entityMap": {},
                        },
                    },
                    "uuid2": {
                        "@type": "custom_type",
                        "searchableText": "custom text foo",
                    },
                },
            },
        )

        self.assertEqual(response.status_code, 204)

        query = {
            "SearchableText": "Volto",
            "metadata_fields": "Title",
            "portal_type": "Document",
        }
        response = self.api_session.get("/@search", params=query)
        json_response = response.json()
        self.assertEqual(json_response["items_total"], 0)

        query = {
            "SearchableText": "Plone",
            "metadata_fields": "Title",
            "portal_type": "Document",
        }
        response = self.api_session.get("/@search", params=query)
        json_response = response.json()
        self.assertEqual(json_response["items_total"], 1)
        self.assertEqual(json_response["items"][0]["Title"], "A document")

        query = {
            "SearchableText": "custom",
            "metadata_fields": "Title",
            "portal_type": "Document",
        }
        response = self.api_session.get("/@search", params=query)
        json_response = response.json()
        self.assertEqual(json_response["items_total"], 1)
        self.assertEqual(json_response["items"][0]["Title"], "A document")

    def test_search_slate_text(self):
        """test_search_text."""
        self.doc.blocks = {
            "38541872-06c2-41c9-8709-37107e597b18": {
                "@type": "slate",
                "plaintext": "Under a new climatic regime, therefore",
                "value": [],
            },
            "4fcfeb9b-f73e-427c-9e06-2e4d53b06865": {
                "@type": "slate",
                "searchableText": "EEA Climate Change data centre",
                "value": [],
            },
        }
        self.doc.blocks_layout = {
            "items": [
                "38541872-06c2-41c9-8709-37107e597b18",
                "4fcfeb9b-f73e-427c-9e06-2e4d53b06865",
            ]
        }
        self.portal.portal_catalog.indexObject(self.doc)

        query = {"SearchableText": "climatic"}
        results = self.portal.portal_catalog.searchResults(**query)
        self.assertEqual(len(results), 1)

        brain = results[0]
        self.assertEqual(brain.Title, "A document")

        query = {"SearchableText": "EEA"}
        results = self.portal.portal_catalog.searchResults(**query)
        self.assertEqual(len(results), 1)

        brain = results[0]
        self.assertEqual(brain.Title, "A document")

    def test_search_table_text(self):
        """Test text in tables is indexed"""
        self.doc.blocks = {
            "uuid1": {
                "@type": "table",
                "table": {
                    "rows": [
                        {
                            "cells": [
                                {
                                    "key": "3dli7",
                                    "type": "header",
                                    "value": {
                                        "blocks": [
                                            {
                                                "key": "1kh23",
                                                "text": "First header",
                                                "type": "unstyled",
                                                "depth": 0,
                                                "inlineStyleRanges": [],
                                                "entityRanges": [],
                                                "data": {},
                                            }
                                        ],
                                        "entityMap": {},
                                    },
                                },
                                {
                                    "key": "3dli8",
                                    "type": "header",
                                    "value": {
                                        "blocks": [
                                            {
                                                "key": "1kh23",
                                                "text": "Second header",
                                                "type": "unstyled",
                                                "depth": 0,
                                                "inlineStyleRanges": [],
                                                "entityRanges": [],
                                                "data": {},
                                            }
                                        ],
                                        "entityMap": {},
                                    },
                                },
                            ],
                        },
                        {
                            "cells": [
                                {
                                    "key": "3dli9",
                                    "type": "data",
                                    "value": {
                                        "blocks": [
                                            {
                                                "key": "1kh23",
                                                "text": "Data foo",
                                                "type": "unstyled",
                                                "depth": 0,
                                                "inlineStyleRanges": [],
                                                "entityRanges": [],
                                                "data": {},
                                            }
                                        ],
                                        "entityMap": {},
                                    },
                                },
                                {
                                    "key": "3dl29",
                                    "type": "data",
                                    "value": {
                                        "blocks": [
                                            {
                                                "key": "1kh23",
                                                "text": "Data bar",
                                                "type": "unstyled",
                                                "depth": 0,
                                                "inlineStyleRanges": [],
                                                "entityRanges": [],
                                                "data": {},
                                            }
                                        ],
                                        "entityMap": {},
                                    },
                                },
                            ],
                        },
                    ],
                },
            }
        }
        self.doc.blocks_layout = {"items": ["uuid1"]}
        self.portal.portal_catalog.indexObject(self.doc)

        query = {"SearchableText": "foo"}
        results = self.portal.portal_catalog.searchResults(**query)
        self.assertEqual(len(results), 1)

        brain = results[0]
        self.assertEqual(brain.Title, "A document")

        query = {"SearchableText": "first"}
        results = self.portal.portal_catalog.searchResults(**query)
        self.assertEqual(len(results), 1)

        brain = results[0]
        self.assertEqual(brain.Title, "A document")



from plone.app.upgrade.utils import loadMigrationProfile
from plone.restapi.testing import PLONE_RESTAPI_DX_INTEGRATION_TESTING
from Products.CMFCore.utils import getToolByName
from unittest import TestCase


class TestUpgrades(TestCase):

    layer = PLONE_RESTAPI_DX_INTEGRATION_TESTING

    def setUp(self):
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]

    def test_migration_profile_to_0002_can_be_loaded(self):
        loadMigrationProfile(self.portal, "profile-plone.restapi.upgrades:0002")
        self.assertTrue(True)

    def test_run_migration_profile_to_0002(self):
        from plone.restapi.upgrades.to0002 import assign_use_api_permission

        portal_setup = getToolByName(self.portal, "portal_setup")
        assign_use_api_permission(portal_setup)
        self.assertTrue(True)

    def test_migration_profile_to_0004_can_be_loaded(self):
        loadMigrationProfile(self.portal, "profile-plone.restapi.upgrades:0004")
        self.assertTrue(True)

    def test_run_migration_profile_to_0004(self):
        from plone.restapi.upgrades.to0004 import assign_get_users_permission

        portal_setup = getToolByName(self.portal, "portal_setup")
        assign_get_users_permission(portal_setup)
        self.assertTrue(True)

    def test_run_migration_profile_to_0005(self):
        from plone.restapi.upgrades.to0005 import rename_tiles_to_blocks

        portal_setup = getToolByName(self.portal, "portal_setup")
        rename_tiles_to_blocks(portal_setup)
        self.assertTrue(True)



from plone.app.querystring.interfaces import IQuerystringRegistryReader
from plone.registry.interfaces import IRegistry
from plone.restapi.services import Service
from zope.component import getMultiAdapter
from zope.component import getUtility


class QuerystringGet(Service):
    """Returns the querystring configuration.

    This basically does the same thing as the '@@querybuilderjsonconfig'
    view from p.a.querystring, but exposes the config via the REST API.
    """

    def reply(self):
        registry = getUtility(IRegistry)
        reader = getMultiAdapter((registry, self.request), IQuerystringRegistryReader)

        result = reader()
        result["@id"] = "%s/@querystring" % self.context.absolute_url()
        return result



from plone.restapi.search.handler import SearchHandler
from plone.restapi.search.utils import unflatten_dotted_dict
from plone.restapi.services import Service


class SearchGet(Service):
    def reply(self):
        query = self.request.form.copy()
        query = unflatten_dotted_dict(query)
        return SearchHandler(self.context, self.request).search(query)



from plone.app.testing import setRoles
from plone.app.testing import SITE_OWNER_NAME
from plone.app.testing import SITE_OWNER_PASSWORD
from plone.app.testing import TEST_USER_ID
from plone.restapi.testing import PLONE_RESTAPI_DX_FUNCTIONAL_TESTING
from plone.restapi.testing import RelativeSession
import unittest


class TestDexterityTypesControlpanel(unittest.TestCase):

    layer = PLONE_RESTAPI_DX_FUNCTIONAL_TESTING

    def setUp(self):
        self.app = self.layer["app"]
        self.portal = self.layer["portal"]
        self.request = self.layer["request"]
        self.portal_url = self.portal.absolute_url()
        setRoles(self.portal, TEST_USER_ID, ["Manager"])

        self.api_session = RelativeSession(self.portal_url, test=self)
        self.api_session.headers.update({"Accept": "application/json"})
        self.api_session.auth = (SITE_OWNER_NAME, SITE_OWNER_PASSWORD)

    def tearDown(self):
        self.api_session.close()

    def test_controlpanels_dexterity_types_get(self):
        response = self.api_session.get("/@controlpanels/dexterity-types")
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            [
                "Collection",
                "Document",
                "Folder",
                "Link",
                "File",
                "Image",
                "News Item",
                "Event",
                "DXTestDocument",
            ],
            [
                x.get("id")
                for x in self.api_session.get("/@controlpanels/dexterity-types")
                .json()
                .get("items")
                if x.get("id") != "Plone Site"
            ],
        )

    def test_controlpanels_dexterity_types_document_get(self):
        response = self.api_session.get("/@controlpanels/dexterity-types/Document")
        self.assertEqual(200, response.status_code)
        self.assertEqual(
            f"{self.portal_url}/@controlpanels/dexterity-types/Document",
            response.json().get("@id"),
        )
        self.assertEqual("Page", response.json().get("title"))

    def test_controlpanels_dexterity_types_post(self):
        response = self.api_session.post(
            "/@controlpanels/dexterity-types",
            json={
                "title": "My Custom Content Type",
                "description": "A custom content-type",
            },
        )

        self.assertEqual(201, response.status_code)
        self.assertEqual(
            "{}/@controlpanels/dexterity-types/my_custom_content_type".format(
                self.portal_url
            ),
            response.json().get("@id"),
        )
        self.assertEqual("My Custom Content Type", response.json().get("title"))
        self.assertEqual("A custom content-type", response.json().get("description"))

    def test_controlpanels_dexterity_types_document_patch(self):
        response = self.api_session.patch(
            "/@controlpanels/dexterity-types/Document",
            json={
                "title": "New Content Type Title",
                "description": "New content type description",
            },
        )

        # PATCH returns no content
        self.assertEqual(204, response.status_code)

        response = self.api_session.get("/@controlpanels/dexterity-types/Document")
        self.assertEqual(200, response.status_code)
        self.assertEqual("New Content Type Title", response.json().get("title"))
        self.assertEqual(
            "New content type description", response.json().get("description")
        )

    def test_controlpanels_dexterity_types_document_delete(self):
        response = self.api_session.delete("/@controlpanels/dexterity-types/Document")

        self.assertEqual(204, response.status_code)
        self.assertEqual(
            [
                "Collection",
                "Folder",
                "Link",
                "File",
                "Image",
                "News Item",
                "Event",
                "DXTestDocument",
            ],
            [
                x.get("id")
                for x in self.api_session.get("/@controlpanels/dexterity-types")
                .json()
                .get("items")
                if x.get("id") != "Plone Site"
            ],
        )



from plone.restapi.services import Service
from plone.restapi.serializer.working_copy import WorkingCopyInfo


class GetWorkingCopy(Service):
    def reply(self):
        baseline, working_copy = WorkingCopyInfo(self.context).get_working_copy_info()

        return {"working_copy": working_copy, "working_copy_of": baseline}
